#!/usr/bin/env python

#  by Jon Tischler (APS/ANL)

import sys
from string import ascii_lowercase, ascii_uppercase, whitespace, digits
import atomGeneral


"""		usage:
from formula import chemFormula
c = chemFormula('TiO2 2(P 3(CO))')
print c.compact
print c.molWt
print c.massSpectrum()
"""

""" ============================================================================
	======================== Start of Chemical Formulas ========================
"""

class chemFormula :
	"""
	Parse a chemical formula into a list (actually a tuple of tuples)
	understands formulas of type:
		"TiO2 2(P 3(CO))".
	Note atomic symbols must have proper Upper/Lower case

	varialbles for this method are:
		self.formula		string version of the formula
		self.compact		a compact string version of the formula
		self.molWt			molecular weight of the molecule
		self.formulaT		tuple version of the formula
		self.compactFormT	compact tuple version of the formula

	also implements the methods:
		self.massSpectrum()		computes the mass spcectrum of the atoms in the formula
								from a list of naturally ocurring isotopes
	"""

	def __init__(self,formula):
		formula = formula.replace('\t',' ')			# change tabs to spaces, clean up the input formula
		formula = formula.replace(',',' ')			# change commas to spaces
		formula = formula.strip()					# remove leading & trailing white space
		while formula.find('  ')>=0:				# all multiple space runs to single spaces
			formula = formula.replace('  ',' ')

		self.formula = formula						# save the cleaned up formula
		self.formulaT = self.__ChemicalFormulaToTuple(formula)
		self.__compact()							# compute and set self.compact & self.compactFormT
		self.__molWt()								# compute and set self.molWt
		return None


	def massSpectrum(self):
		"""
		Returns spectrum of all isotopes in the with the relative strength of each
		if uses a list of naturally ocurring isotopes.
		"""
		unNormalizedSpectrum = tuple()
		fMax = 0.0
		for symb,num in self.compactFormT:
			try:
				isotopTuples = atomGeneral.isotope(symb).isotopeListNatural()	# list of naturally occurring isotopes
			except:
				return None

			for amu,frac in isotopTuples:
				frac *= num
				fMax = max(frac,fMax)					# find the maximum peak
				unNormalizedSpectrum += ((amu,frac),)

		spectrum = tuple()								# finally normalize so max peak is 1.0
		for amu,num in unNormalizedSpectrum:
			spectrum += ((amu,num/fMax),)				# normalize each mass so peak is 1.0
		return spectrum

	"""	========== the following procedures are only called internally ========== """

	def __repr__(self):
		return 'chemFormula[%r,  compact=%r,  molWt=%r]' % (self.formulaT, self.compactFormT, self.molWt)


	def __str__(self):
		return 'chemFormula("%s",  compact="%s",  molWt=%g)' % (self.formula, self.compact, self.molWt)


	def __molWt(self):
		""" Returns molecular weight of the molecule, also sets self.molWt """
		self.molWt = float(0.0)
		for symb,num in self.compactFormT:
			try:
				self.molWt += num * atomGeneral.baseAtom(symb).amu
			except:
				return None


	def __compact(self):
		"""
		Combine all occurances of each atom into one occurance
		this sets the internal compact form to self.compactFormT, but it
		returns the compact form string self.compactForm.
		"""

		atoms = list()
		nums = list()
		for item in self.formulaT:
			atom = item[0]
			num = item[1]
			try:
				a = atoms.index(atom)
				num += nums[a]
				nums[a] = num
			except:
				atoms += (atom,)
				nums += (num,)
			
		self.compactFormT = tuple()
		for atom,num in zip(atoms,nums):
			self.compactFormT += ((atom,num),)

		self.compact = self.__list2str(self.compactFormT)	# returns the compact form as STRING


	def __ChemicalFormulaToTuple(self,formula):
		"""
		Parse a chemical formula into a list (actually a tuple of tuples)
		understands formulas of type:
			"TiO2 2(P 3(CO))".
		Note atomic symbols must have proper Upper/Lower case
		formula		a str, chemical formula, Elements start with a capital letter.
		"""

		formula = formula.replace('\t',' ')			# change tabs to spaces
		formula = formula.replace(',',' ')			# change commas to spaces
		formula = formula.strip()					# remove leading & trailing white space
		while formula.find('  ')>=0:				# all multiple space runs to single spaces
			formula = formula.replace('  ',' ')

		N = len(formula)
		i = 0
		out = tuple()
		while i<N:									# over each character in formula
			try:	c = formula[i]
			except:	break							# all done, no more ascii

			if c==' ':								# skip spaces
				i += 1
				continue

			(multiplier,numLen) = self.__leadingNumber(formula[i:])
			if numLen>0:							# multiplier with a sub formula, sub formulas MUST have a '(...)'
				i0 = formula.find("(",i)
				i1 = self.__FindMatchingBracket(formula,i0)
				if (i1<=i0):
					print 'ERROR -- ChemicalFormulaToTuple(), cannot find matching parenthsis in  "'+formula+'"'
					return ""
				subList = self.__ChemicalFormulaToTuple(formula[i0+1:i1])
				subList = self.__multiplyElemetnList(subList,multiplier)	# increment subList by multiplier
				out += subList
				i = i1+1
				continue

			if not(c in ascii_uppercase): 				# formula is empty, all done (all elements start with a capital letter)
				print 'ERROR -- ChemicalFormulaToTuple(), atomic symbols must start with a CAPITAL letter, could not parse "'+formula+'"'
				return ""

			symb = formula[i]
			i += 1										# move pointer ahead one

			try:	c = formula[i]
			except:	c = ''

			if c in ascii_lowercase:					# next char is a lower case letter, make it part of symb
				symb += c								# optional second character of atomic symbol
				i += 1									# move pointer ahead one

			try:	c = formula[i]						# after symbol will either be a number or another symbol, or end of string
			except:	c = ''

			if c in ascii_uppercase:					# found a capital letter
				num = 1
			elif i>=N:									# found end of string
				num = 1
			elif c in whitespace:						# found whitespace, skip
				num = 1
				i += 1
			else:
				(num,numLen) = self.__leadingNumber(formula[i:N])	# if it is not letter or white, it must be a number
				if numLen<1:
					print 'ERROR -- ChemicalFormulaToTuple(), found invalid number, could not parse "'+formula+'"'
					return ''
				i += numLen								# got the number after a symbol, find char after the number
			out += ((symb,num),)
		return out


	def __list2str(self,atomList):
		""" Converts the long list form into a string form, it does no compaction. """
		out = str()
		for atom,num in atomList:
			if num==1:	out += atom
			elif num>0:	out += atom+('%g' % num)
		return out


	def __leadingNumber(self,s):
		"""
		Find float value of leading number in a string, 
		If there string does not start with a number, then it returns None
		when the string starts with a number it returns a tuple with the value
		and the lenght of string that made up the number
		"""
		if not(type(s) is str): return (None,0)			# only works on a string
		ss = ''
		for c in s:
			if c in digits+'+.':
				ss += c
			else:
				break

		try:	val = float(ss)
		except:	val = None

		if type(val) is float: return (val,len(ss))
		return (None,0)


	def __multiplyElemetnList(self,inList,mult):
		"""
		inList	a tuple or list
		mult	a number
		"""

		out = tuple()
		if not(type(inList) is tuple): return out	# must be a tuple or list
		if len(inList)<1: return out
		try:
			mult = float(mult)			# mult must be a number
			if not(mult==mult): return out
		except:
			return out

		for item in inList:
			try:
				symb = item[0]
				num = item[1]*mult
			except:
				return out
			out += ((symb,num),)

		return out


	def __FindMatchingBracket(self,ss,istart):
		"""
		ss		string to search
		istart	location of b0 in ss
		"""
		bOpen="([{"				# all allowed open brackets
		bClose=")]}"			# all allowed close brackets
		b0 = ss[istart]			# starting bracket, can be "(" "[", or "{"

		ib = bOpen.find(b0)		# identifies the type of opening bracket
		if ib<0: return -1		# istart is not on a valid bracket
		b1=bClose[ib]			# b1 is the closing bracket
		while (True):
			iclose = ss.find(b1,istart)
			if ss.find(b0,istart+1,iclose) < 1: break
			istart = iclose+1

		return iclose


""" ========================= End of Chemical Formulas =========================
	============================================================================
"""

if __name__ == '__main__':
	"""
	Main function for formula.py.
	Test cases for for class chemFormula to verify correct behavior.
	"""
	try:	testGroup = int(sys.argv[1])
	except:	raise ValueError('./chemFormula testGroup,  testGroup must be a bit mask, a combo of 1-1024, use -1 for all')
	print 'showing testGroup(s) == %r (= %r)' % (testGroup,bin(testGroup))

	def test_chemFormula(formula):
		c = chemFormula(formula)
		print '\ntest of formula: %r' % formula
		print 'str --> ',str(c)
		print 'repr --> ',repr(c)
		print 'massSpectrum = ',c.massSpectrum()

	print '\nclass chemFormula: ',chemFormula.__doc__

	if testGroup & 1:	test_chemFormula('V')
	if testGroup & 2:	test_chemFormula('V2')
	if testGroup & 4:	test_chemFormula('Fe2O3')
	if testGroup & 8:	test_chemFormula('TiO2 2(P)')
	if testGroup & 16:	test_chemFormula('TiO2 2(P 3(CO))')
