#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Lattice.py
#
# $Id:    $
# $URL: $
#
# Part of the "pydiffract" package
#

import os
import sys
import math
import cmath
import string
import xml.dom.minidom
from fnmatch import fnmatch
import numpy as np
from JZTutil import ConvertTemperatureUnits, ConvertUnits2meters, ConvertAngleUnits
from symrange import symrange
import atomGeneral

__version__ = "$Revision: $"
__author__  = "Jon Z. Tischler, <tischler@aps.anl.gov>" +\
              "Argonne National Laboratory"
__date__    = "$Date: $"
__id__      = "$Id: $"


SHOW_ALL = True

NaN = float('nan')
amu_eV = 931.4940954e6				# energy of one amu (eV),  these 4 numbers only used by Debye-Waller calculation
hbar = 6.582119514e-16				# h/2PI (eV-sec)
kB = 8.6173303e-5					# Boltzmann constant (eV/K)
c_ms = 299792458					# speed of light (m/sec)
NA = 6.022140857e23					# Avagadro's number
Zmax = 109							# maximum atomic number



#pragma ModuleName=LatticeSym
#include "MaterialsLocate"								// used to find the path to the materials files


"""
	Rhombohedral Transformation:

	for Rhombohedral (hkl), and Hexagonal (HKL)
		h = (2H+K+L)/3
		k = (-H+K+L)/3
		l = (-H-2K+L)/3
		and the condition (-H+K+L) = 3n (where n is an integer) for allowed reflections, three Rhombohedral cells per Hexagonal cell

		H = h=k
		K = k-l
		L = h+k+l

	for Hexagonal lattice constants aH,cH (alpha=beta=90, gamma=120)
	a(Rhom) = sqrt(3*aH^2 + cH^2)/3
	sin(alpha(Rhom)/2) = 1.5/sqrt(3+(cH/aH)^2)
	These equations are implented in the routines:  Hex2Rhom(aH,cH)  &  Rhom2Hex(aR,alpha)
"""


class atomXtal(atomGeneral.CromerAtom):
	""" A Class that contains the information about the atom in a crystal.
		If you give it one atom position (x,y,z) and the symmetry operations (symOps),
		then it will also calculate and store all of the symmetry equivalent atom 
		positions for this crystal.
		It will also hold the thermal parameters for the atom, and can return atomic structure factor given a Q
	"""
	# note, atomGeneral.CromerAtom is only neede to provide f0, if you do not include it, the fatom method will still work

	def __init__(self, label, x,y,z, Zatom=-1, valence=0, occ=1.0, WyckoffSymbol='', symOps=None, mult=0,
		DebyeT=NaN,Biso=NaN,Uiso=NaN,U11=NaN,U22=NaN,U33=NaN,U12=NaN,U13=NaN,U23=NaN):
		""" Initialize the atomXtal instance. """
		try:	Zatom = int(Zatom)
		except:	Zatom = -1
		if not(Zatom>=1 and Zatom<=Zmax):	# try to get Zatom from label
			Zatom = ZfromLabel(label)
			if Zatom < 1: raise ValueError('Cannot determine Z of atom from "%s"' % label)

		atomGeneral.CromerAtom.__init__(self, Zatom)	# sets self.Z, self.sym 

		self.label = str(label)			# label for this atom, usually starts with atomic symbol
		self.x = float(x)				# fractional coordinates along lattice vectors
		self.y = float(y)
		self.z = float(z)
		self.occ = float(occ)			# occupancy of this atom

		try:	self.valence = int(valence)
		except:	raise ValueError('valence must be a simple integer, not "%r"' % valence)
		if abs(self.valence)>10:
			raise ValueError('valence of %d is too big, it must be < 10' % self.valence)

		if type(WyckoffSymbol) is unicode: WyckoffSymbol = str(WyckoffSymbol)
		self.WyckoffSymbol = WyckoffSymbol	# a single letter, e.g. 'a', also called the Wyckoff letter
		self.DebyeT = float(DebyeT)		# Debye Temperature (K),  for DebyeT, B, Uiso, & U_ij, use only one method
		self.Biso = float(Biso)			# B-factor (nm^2) using:   exp(-M) = exp(-B * sin^2(theta)/lam^2)		B = 8 * PI^2 * <u^2> =  8 * PI^2 * Uiso,	exp[-B*q^2 / (16 PI^2) ]
		self.Uiso = float(Uiso)			# isotropic U (nm^2)
		self.U11 = float(U11)			# anisotropic U(11) (nm^2)
		self.U22 = float(U22)			# anisotropic U(22) (nm^2)
		self.U33 = float(U33)			# anisotropic U(33) (nm^2)
		self.U12 = float(U12)			# anisotropic U(12) (nm^2)
		self.U13 = float(U13)			# anisotropic U(13) (nm^2)
		self.U23 = float(U23)			# anisotropic U(23) (nm^2)
		self.SetHasThermalInfo()		# True if atom has DebyeT, Biso, Uiso, or Uij
		if self.bad(): raise ValueError('atom values are invalid')
		self.xyz = None					# only used with the lattice
		try:	self.mult = int(mult)
		except:	self.mult = 0
		self.mult = max(1,self.mult)	# got to have at least 1

		if not(symOps is None):			# calculate all atom positions from symOps, this can be done later
			self.calcAllAtomPositions(symOps)

		return None


	def __str__(self):
		""" Return string value for atomXtal. """
		if self.valence==0:	vstr = ''
		else:				vstr = ', v=%+d' % self.valence
		if len(self.WyckoffSymbol):
			out = "     %s (Z=%g%s)\t%s\t{%g,  %g,  %g}" % (self.label,self.Z,vstr,self.WyckoffSymbol,self.x,self.y,self.z)
		else:
			out = "     %s (Z=%g%s)\t\t{%g,  %g,  %g}" % (self.label,self.Z,vstr,self.x,self.y,self.z)

		if self.occ != 1:	out += "    occ = %g" % self.occ
		if self.mult > 1:	out += "\tmultiplicity = %d" % self.mult

		# Thermal vibration information, print at most one kind of info
		if self.DebyeT > 0:	out += "    Debye Temperature = %g K" % self.DebyeT
		elif self.Biso > 0:	out += "    Isotropic B = %g (Angstrom^2)" % (self.Biso * 100)
		elif self.Uiso > 0:	out += "    Isotropic U = %g (Angstrom^2)" % (self.Uiso * 100)
		elif self.U11 > 0 or self.U22 > 0 or self.U33 > 0:
			out += "    An-Isotropic U, U11 = %g, U22 = %g, U33 = %g (nm^2)" % ((self.U11 *100),(self.U22 *100),(self.U33 *100))
			if self.U12 > 0 or self.U13 > 0 or self.U23 > 0:
				out += "    An-Isotropic U, U12 = %g, U13 = %g, U23 = %g (nm^2)" % ((self.U12 *100),(self.U13 *100),(self.U23 *100))

		if self.xyz:
			i = 0
			for xyz in self.xyz:
				(x,y,z) = xyz
				out += '\n\t%d\t{%g,  %g,  %g}' % (i+1,x,y,z)
				i += 1

		return out


	def __repr__(self):
		""" Return representation value for atomXtal. """
		out = 'atomXtal[%r, %r, Z=%r, valence=%r, xyz=(%r, %r, %r)' % (self.label,self.sym,self.Z,self.valence,self.x,self.y,self.z)
		if len(self.WyckoffSymbol): out += ', Wyckoff=%r' % self.WyckoffSymbol
		if not (self.occ == 1): out += ', occ=%r' % self.oc
		if not (self.mult == 1): out += ', mult=%r' % self.mult

		# Thermal vibration information, print at most one kind of info
		if self.DebyeT > 0:	out += ', DebyeT=%r' % self.DebyeT
		elif self.Biso > 0:	out += ', Biso= %r' % self.Biso
		elif self.Uiso > 0:	out += ', Uiso= %r' % self.Uiso
		elif self.U11 > 0 or self.U22 > 0 or self.U33 > 0:
			out += ', U11=%g, U22=%r, U33=%r' % (self.U11,self.U22,self.U33)
			if self.U12 > 0 or self.U13 > 0 or self.U23 > 0:
				out += ', U12=%r, U13=%r, U23=%r' % (self.U12,self.U13,self.U23)

		out += ']'
		if self.xyz:
			i = 0
			for xyz in self.xyz:
				(x,y,z) = xyz
				out += '\n\t%d\t{%g,  %g,  %g}' % (i+1,x,y,z)
				i += 1
		return out


	def bad(self):
		if len(self.label)<1: return True
		if not (self.x >= 0 and self.y >= 0 and self.z >= 0): return True
		if not (self.occ > 0 and self.occ <= 1): return True
		if not (type(self.Z) is int or type(self.Z) is long): return True
		if not (self.Z >=1 and self.Z <= Zmax): return True
		if not (type(self.WyckoffSymbol) is str): return True
		if len(self.WyckoffSymbol)>1: return True

		if self.DebyeT < 0: return True			# no negatives, but NaN is OK
		if self.Biso < 0: return True
		if self.Uiso < 0: return True
		if self.U11 < 0: return True
		if self.U22 < 0: return True
		if self.U33 < 0: return True
		if self.U12 < 0: return True
		if self.U13 < 0: return True
		if self.U23 < 0: return True

		usingT = 0
		if self.DebyeT > 0: usingT += 1
		if self.Biso > 0: usingT += 1
		if self.Uiso > 0: usingT += 1
		if (self.U11 > 0 or self.U22 > 0 or self.U33 > 0): usingT += 1
		if usingT>1: return True				# too many thermal vibration methods
		if self.U11 > 0:						# using anisotropic U
			if not(self.U22 >=0 ): return True	# if U11 is used, all symmetric U's must be valid
			if not(self.U33 >=0 ): return True
			if self.U12 > 0:					# using anisotropic U cross-terms
				if not(self.U13 >=0 ): return True
				if not(self.U23 >=0 ): return True

		return False							# all OK


	def SetHasThermalInfo(self):
		""" sets the kind of thermal/vibrational info present in this atom """
		Tinfo = Vinfo = False
		if self.DebyeT > 0: Tinfo = True
		elif self.Biso > 0: Vinfo = True
		elif self.Uiso > 0: Vinfo = True
		elif self.Biso > 0: Vinfo = True
		elif (self.U11 > 0 and self.U22 > 0 and self.U33 > 0): Vinfo = True
		self.hasThermalInfo = Vinfo or Tinfo	# True if atom has DebyeT, Biso, Uiso, or Uij
		self.Tinfo = Tinfo
		self.Vinfo = Vinfo


	def DW_factor_M(self,T,thetaM,Q):
		""" calculates the M in exp(-M), no I/O
		This is the x-ray Debye-Waller factor, that's what the '_M' means.
		T			# Temperature (K)
		thetaM		# Debye Temperature (K)
		Q			# length of q vector (1/nm)
		amu			# mass of atom (amu)
		"""
		if T<0 or thetaM<0 or Q<0 or self.amu<0: return float('nan')
		elif T<=0:	return 0.0
		xx = float(thetaM) / float(T)
		if xx>50000.0: return 0.0
		Phi = simpsonIntegral(self.PhiIntegrand,0,xx).calc() / xx
		B = (3*hbar*hbar * T * c_ms*c_ms)/(2*(self.amu*amu_eV)*kB*thetaM*thetaM)* (Phi + xx/4)
		M = B *(Q*Q*1.e18)						# Q is in 1/nm, but we need it in 1/m
		return M

	def PhiIntegrand(self,xx):
		""" this is the function we need to integrate from DW_factor_M()
		this function should never be called with xx<0
		"""
		if xx>0:	return xx / ( math.exp(xx) - 1 )
		else:		return 1.0


	def calcAllAtomPositions(self,symOps):
		""" calculate all the atom positions from the symOps
		symOps are the symmetry operations for this atom, supplied by calling function
		symOps are a set of equivXYZ1
		results are stored in self.xyz
		self.mult is also set here
		"""

		self.xyz = list()					# init to empty list
		xyz1 = np.matrix([self.x, self.y, self.z, 1.0])
		for symOp in symOps:
			symOp = np.matrix(symOp)
			xyzTest = symOp * xyz1.T
			x = float(xyzTest[0])
			y = float(xyzTest[1])
			z = float(xyzTest[2])
			x += abs(math.floor(x))			# translate back in to unit cell, so value in [0,1)
			x = x % 1.
			y += abs(math.floor(y))
			y = y % 1.
			z += abs(math.floor(z))
			z = z % 1.

			dup = False
			for xyzHave in self.xyz:		# check for duplicates
				if (abs(xyzHave[0]-x) + abs(xyzHave[1]-y) + abs(xyzHave[2]-z))<1e-3:
					dup = True
					break

			if not dup:						# not a duplicate, so add to the list of positions
				self.xyz.append((x,y,z))

		self.mult = len(self.xyz)


	def fatom(self,Q,keV):
		""" This Cromer & Liberman values
		Q		|Q| (1/nm)
		keV		energy (keV)		Do NOT yet have energy dependence
		Note, if CromerAtom.f0 has NOT been imported, then this method still works.
		"""
		try:	zf0 = self.f0(Q)
		except:	zf0 = float(max(self.Z - self.valence,0))
		return complex(zf0,0.0)




class bondType(object):
	""" A Class that the definition of a bond in a crystal.
		It store the label of the atoms at each end, and the bond length.
		The bond length (lengths) can be a tuple if more than one length is possible.
	"""

	def __init__(self, label0,label1,lengths, btype=1):
		"""
		Initialize the bondType instance.
		this bond goes between atoms label0 and label1 with a length of lengths

		label0			label for first atom, usually starts with atomic symbol, e.g. Si001
		label1			label for second atom, usually starts with atomic symbol
		lengths			tuple with bond lengths (nm) e.g. (1,) or 0.2 or (1.0, 1.5)
		btype			optional bond type (e.g. 1 for single bonds, 2 for double bonds ...)
		"""

		if not(type(label0) is str or type(label0) is unicode):
			raise ValueError('bond label0 is not a string')
		if not(type(label1) is str or type(label1) is unicode):
			raise ValueError('bond label1 is not a string')
		if len(label0) < 1:
			raise ValueError('bond label0 is an empty string')
		if len(label1) < 1:
			raise ValueError('bond label1 is an empty string')
		self.label0 = label0			# label for first atom, usually starts with atomic symbol
		self.label1 = label1			# label for second atom, usually starts with atomic symbol

		if type(lengths) is int or type(lengths) is float:
			lengths = (float(lengths),)	# a single number was entered, make it a tuple of floats
		self.lengths = tuple()			# length of bond (possibly multiple values) (nm)
		for l in lengths:
			if type(l) is float: self.lengths += (l,)
			else: raise ValueError('bond length is not a float')

		if len(self.lengths) < 1:
			raise ValueError('there is no bond lengths')

		try:
			self.btype = int(btype)		# bond type is a positive int > 0
		except:
			raise ValueError('bond type is not a positive integer, btype = %r' % btype)
		if self.btype<1: raise ValueError('bond multiplicity is not a positive integer, btype = %r' % btype)


	def __str__(self):
		""" Return string value for bondType. """
		lenStr = ''
		for l in self.lengths:
			if len(lenStr): lenStr += ', '
			lenStr += "%.4g" % l

		try:
			sbtype = ['single','double','triple','quadruple','quintuple'][self.btype - 1]
		except:
			sbtype = '%r-tuple' % self.btype
		return "     (%s) <--> (%s):  %s nm,  %s bond" % (self.label0,self.label1,lenStr,sbtype)


	def __repr__(self):
		""" Return printable representation for bondType. """
		return 'bondType[%s, %s, %r, %d]' % (self.label0,self.label1,self.lengths,self.btype)





class Lattice(object):
	""" A Class that the defines a crystal lattice with all of its atoms.
		It can load all the information from a file
		All parameters are forced to be consistent with the space group number.
		It can also calculate the structure factor F(hkl)
	"""

	def __init__(self, SpaceGroup, a,b,c,alpha,beta,gam, desc='', Temperature=NaN,alphaT=None,atoms=tuple(),bonds=tuple(),file='',allowed_F_N=0.01):
		"""
		Initialize the Lattice instance.

		values that are passed to the __init__()
		self.desc = desc					# name or decription of this crystal
		self.SpaceGroup = SpaceGroup		# Space Group number from international tables, allowed range is [1, 230]
		self.a = float(a)					# lattice constant (nm)
		self.b = float(b)
		self.c = float(c)
		self.alpha = float(alpha)			# angles (degree)
		self.beta = float(beta)
		self.gam = float(gam)
		self.alphaT = alphaT				# optional: coef of thermal expansion, a = ao*(1+alphaT*(TempC-22.5))
		self.Temperature = Temperature		# optional: temperature for theses values, defaul is 22.5 C
		self.atoms							# optional: a tuple of atomXtal's (no atoms are required)
		self.bonds							# optional: a tuple of bondType's (no bonds are required)
		self.allowed_F_N					# optional: number of electrons/atom (e.g. |F|/Natmos) for an allowed reflection, (default=0.01)

		the following values are calculated at __init__()
		self.latticeSystemName				# Cubic, Hexagonal, Trigonal, Tetragonal, Orthorhombic, Monoclinic, Triclinic
		self.density						# calculated density (g/cm^3)
		self.formulaMin						# calculated minimal chemical formula for this structure
		self.direct							# calculated direct lattice (nm)
		self.recip							# calculated reciprocal lattice (1/nm), has the 2pi in it

		important methods:
		def Hex2Rhom(self,aH,cH)			# convert Hexagonal lattice constants to Rhombohedral
		def Rhom2Hex(self,aR,alpha)			# convert Rhombohedral lattice constants to Hexagonal
		def dSpacing(self,h,k,l,T=None)		# returns d-spacing for the hkl (nm) 
		def hkl2Q(self,h,k,l)				# compute qvector for (h,k,l) (nm^1) contains the 2*PI
		def uvw2ZoneDir(self,u,v,w)			# compute zone direction of vector for (u,v,w)
		def angleBetweenHKLs(self,h1,k1,l1, h2,k2,l2)	# find angle between (h1,k1,l1) and (h2,k2,l2)
		def angleBetweenZones(self,u1,v1,w1, u2,v2,w2)	# find angle between two zones: (u1,v1,w1) and (u2,v2,w2)
		def findClosestHKL(self,dIN)		# Find the hkl(s) closest to given d-spacing (nm)
		def Fstruct(self,h,k,l,keV=None,T_K=None)	# returns the complex structure factor, (not yet includes T & KeV)
		def allowedHKL(self,h,k,l):			# True if hkl is allowed
		"""

		try:	self.allowed_F_N = float(allowed_F_N)# an allowed reflection has |F|/Natoms > ALLOWED_F_N, must have at least 0.01 electrons/atom
		except: self.allowed_F_N = 0.01

		if len(file)>0:
			self.parseCIFxml(file)

		else:
			self.SpaceGroup = SpaceGroup			# Space Group number from international tables, allowed range is [1, 230]
			if not self.__validSpaceGroup():
				raise ValueError('SpaceGroup must be an integer in range [1,230], not "%r".' % SpaceGroup)
			self.SpaceGroup = int(SpaceGroup)
			self.desc = desc						# name or decription of this crystal
			self.a = float(a)						# lattice constant (nm)
			self.b = float(b)
			self.c = float(c)
			self.alpha = float(alpha)				# angles (degree)
			self.beta = float(beta)
			self.gam = float(gam)

			self.alphaT = alphaT
			self.Temperature = Temperature
			if not (Temperature is None):
				if Temperature < -273.15: raise ValueError("Temperature (C) is %r, which must be > -273.15" % Temperature)

			self.atoms = atoms
			self.bonds = bonds

			self.SetSymmetryOperations()			# these are needed to calculate the atom positions
			self.__ForceLatticeToStructure()		# force a,b,c, alpha,beta,gam to match Space Group, sets Vc, direct & recip too
			self.__setDirectRecip()					# set direct & reicp lattices, also calc Vc

			# set miscelaneous things in lattice
			self.Vibrate = self.haveDebyeT = False
			for atom in self.atoms:
				if atom.hasThermalInfo: self.Vibrate = True	# True if some Thermal vibration info present in xtal (for any atom) 
				if atom.DebyeT > 0: self.haveDebyeT = True	# True if some one of the atoms has a Debye Temperature

			for atom in self.atoms:					# check that all atoms are valid
				if len(self.equivXYZ1): atom.calcAllAtomPositions(self.equivXYZ1)	# have the symmetry operation matricies, use them
			self.__calcDensity()					# now that I have all of the atom positions, can calc density
			self.__MinimalChemFormula()				# set the minimal chemical formula

			# done setting, now check that everything is valid
			if not (self.a >= 0 and self.b >= 0 and self.c >= 0): raise ValueError('invalid a=%r, b=%r, c=%r' % (self.a,self.b,self.c))
			if not (self.alpha > 0 and self.beta > 0 and self.gam > 0): raise ValueError('invalid alpha=%r, beta=%r, gamma=%r' % (self.alpha,self.beta,self.gamma))
			if not (self.alpha < 180 and self.beta < 180 and self.gam < 180): raise ValueError('invalid alpha=%r, beta=%r, gamma=%r' % (self.alpha,self.beta,self.gamma))
			if not self.__validSpaceGroup(): raise ValueError('invalid SpaceGroup = %r' % self.SpaceGroup)
			if (self.Vibrate or self.haveDebyeT) and len(self.atoms) < 1:
				raise ValueError("Inavlid, have Vibrate or DebyeT, but no atoms")

			self.latticeSystemName = self.__latticeSystemName()

#			Tinfo = Vinfo = False					# check for consistent Thermal/Vibrational info
#			for atom in self.atoms:
#				if atom.bad(): raise ValueError('bad atom')
#				Tinfo = Tinfo or atom.Tinfo
#				Vinfo = Vinfo or atom.Vinfo
#			if Tinfo and Vinfo: raise ValueError('there are some atoms Vibrational info and some with Debye Temperature, cannot mix')

			labels = list()							# check for duplicate atoms labels
			for atom in self.atoms:
				labels.append(atom.label)
			if len(set(labels))<len(labels): raise ValueError('there are atoms with duplicate labels')
	
			#		double Unconventional00,Unconventional01,Unconventional02	// transform matrix for an unconventional unit cel
			#		double Unconventional10,Unconventional11,Unconventional12
			#		double Unconventional20,Unconventional21,Unconventional22
			#		char hashID[HASHID_LEN]	// hash function for this strucutre (needs to hold at least 64 chars), This MUST be the last item

		if not self.__isValid(): raise ValueError('Lattice constants are invalid:  a=%g, b=%g, c=%g, alpha=%g, beta=%g, gamma=%g' % (self.a,self.b,self.c, self.alpha,self.beta,self.gam))




	def __validSpaceGroup(self):
		""" must be an integer or long in range [1,230] """
		SG = self.SpaceGroup
		if not ( (type(SG) is int) or (type(SG) is long) ):	return False
		if SG<1:  return False
		if SG>230: return False
		return True


	def __isValid(self):
		""" Check if lattice constants are valid, returns True if all 6 are valid """
		if not self.__validSpaceGroup(): return False

		SG = self.SpaceGroup			# local value for convienence
		a = self.a
		b = self.b
		c = self.c
		alpha = self.alpha
		beta = self.beta
		gam = self.gam

		if math.isnan(a+b+c+alpha+beta+gam): return False
		if not (a>0 and b>0 and c>0 and alpha>0 and beta>0 and gam>0): return False
		if not (a<180 and b<180 and c<180): return False

		if SG >= 195:					# Cubic
			if not (a == b and a == c): return False
			if not (alpha==90.0 and beta==90.0 and gam==90.): return False
		elif SG >= 168:					# Hexagonal
			if not (a == b): return False
			if not (alpha==90.0 and beta==90.0 and gam==120.): return False
		elif SG >= 143:					# Trigonal (generally hexagonal cell), for rhomohedral use rhomohedral cell, unless obviously the hexagonal
			if (isRhombohedral(SG)):# a possible rhombohedral structure
				if ((abs(90-self.alpha)+abs(90-self.beta)+abs(120-self.gam))<1e-6):	# obviously hexagonal
					self.b = self.a
					self.alpha = self.beta = 90
					self.gam = 120
				else:					# rhombohedral with rhombohedral cell
					self.b = self.c = self.a
					self.beta = self.gam = self.alpha
			else:						# "P" trigonal, use hexagonal cell
				self.b = self.a
				self.alpha = self.beta = 90
				self.gam = 120

		elif SG >= 75:					# Tetragonal
			if not (a == b): return False
			if not (alpha==90.0 and beta==90.0 and gam==90.): return False
		elif SG >= 16:					# Orthorhombic
			if not (alpha==90.0 and beta==90.0 and gam==90.): return False
		elif SG >= 3:					# Monoclinic
			if not (alpha==90.0 and gam==90.): return False
		else:							# Triclinic
			pass

		return True


	def __ForceLatticeToStructure(self):
		""" forces lattice constants to match the Space Group number (e.g. for cubic, forces b and c to be a, and all angles 90)
		Cubic			[195,230]		a
		Hexagonal		[168,194]		a,c
		Trigonal		[143,167]		a,alpha
		Tetragonal		[75,142]		a,c
		Orthorhombic	[16,74]			a,b,c
		Monoclinic		[3,15]			a,b,c,gamma
		Triclinic		[1,2]			a,b,c,alpha,beta,gamma
		"""
		SG = self.SpaceGroup			# local value for convienence
		if SG >= 195:					# Cubic
			self.b = self.c = self.a
			self.alpha = self.beta = self.gam = 90
		elif SG >= 168:					# Hexagonal
			self.b = self.a
			self.alpha = self.beta = 90
			self.gam = 120
		elif SG >= 143:					# Trigonal (generally hexagonal cell), for rhomohedral use rhomohedral cell, unless obviously the hexagonal
			if (isRhombohedral(SG)):# a possible rhombohedral structure
				if ((abs(90-self.alpha)+abs(90-self.beta)+abs(120-self.gam))<1e-6):	# obviously hexagonal
					self.b = self.a
					self.alpha = self.beta = 90
					self.gam = 120
				else:					# rhombohedral with rhombohedral cell
					self.b = self.c = self.a
					self.beta = self.gam = self.alpha
			else:						# "P" trigonal, use hexagonal cell
				self.b = self.a
				self.alpha = self.beta = 90
				self.gam = 120

		elif SG >= 75:					# Tetragonal
			self.b = self.a
			self.alpha = self.beta = self.gam = 90
		elif SG >= 16:					# Orthorhombic
			self.alpha = self.beta = self.gam = 90
		elif SG >= 3:					# Monoclinic
			self.alpha = self.gam = 90
		else:							# Triclinic
			pass

		if not(self.a > 0 or self.b > 0 or self.c > 0): 
			raise ValueError('Lattice constants must be >= 0, a=%g, b=%g, c=%g' % (self.a,self.b,self.c))
		elif not ((self.alpha > 0 and self.alpha < 180) or (self.beta > 0 and self.beta < 180) or (self.gam > 0 and self.gam < 180)):
			raise ValueError('Lattice angles must be in (0,180 degree), alpha=%g, beta=%g, gamma=%g' % (self.alpha,self.beta,self.gam))


	def __setDirectRecip(self):
		""" set direct and recip lattice vectors from {a,b,c,alpha,beta,gamma}, also calculates Vc """
		a = self.a
		b = self.b
		c = self.c
		sa = math.sin((self.alpha)* math.pi / 180.)
		ca = math.cos((self.alpha)* math.pi /180.)
		cb = math.cos((self.beta)* math.pi /180.)
		cg = math.cos((self.gam)* math.pi /180.)

		phi = math.sqrt(1.0 - ca*ca - cb*cb - cg*cg + 2*ca*cb*cg)	# = Vc/(a*b*c)
		self.Vc = a*b*c * phi					# volume of unit cell
		pv = (2*math.pi) / (self.Vc)			# used for scaling

		# components of the direct lattice vectors
		a0 = a*phi/sa	; a1 = a*(cg-ca*cb)/sa	; a2 = a*cb
		b0 = 0.0		; b1 = b*sa				; b2 = b*ca
		c0 = 0.0		; c1 = 0.0				; c2 = c
		a0 = zeroOut(a0) ; a1 = zeroOut(a1) ; a2 = zeroOut(a2)
		b1 = zeroOut(b1) ; b2 = zeroOut(b2)
		c2 = zeroOut(c2)
		self.direct = np.matrix([ [a0,b0,c0], [a1,b1,c1], [a2,b2,c2] ])

		as0 = zeroOut((b1*c2-b2*c1)*pv)			# (b x c)*2PI/Vc
		as1 = zeroOut((b2*c0-b0*c2)*pv)
		as2 = zeroOut((b0*c1-b1*c0)*pv)
		bs0 = zeroOut((c1*a2-c2*a1)*pv)			# (c x a)*2PI/Vc
		bs1 = zeroOut((c2*a0-c0*a2)*pv)
		bs2 = zeroOut((c0*a1-c1*a0)*pv)
		cs0 = zeroOut((a1*b2-a2*b1)*pv)			# (a x b)*2PI/Vc
		cs1 = zeroOut((a2*b0-a0*b2)*pv)
		cs2 = zeroOut((a0*b1-a1*b0)*pv)
		self.recip = np.matrix([ [as0, bs0, cs0], [as1, bs1, cs1], [as2, bs2, cs2] ] )
		#
		#	Variable allZero = abs(xtal.Unconventional00)+abs(xtal.Unconventional01)+abs(xtal.Unconventional02)
		#	allZero += abs(xtal.Unconventional10)+abs(xtal.Unconventional11)+abs(xtal.Unconventional12)
		#	allZero += abs(xtal.Unconventional20)+abs(xtal.Unconventional21)+abs(xtal.Unconventional22)
		#	xtal.Unconventional00 = allZero==0 ? NaN : xtal.Unconventional00
		#	if (numtype(xtal.Unconventional00)==0 && xtal.Unconventional00>-100 && xtal.Unconventional00<100)
		#		Make/N=(3,3)/O/D root:Packages:Lattices:Unconventional
		#		Wave Unconventional=root:Packages:Lattices:Unconventional
		#		Unconventional[0][0]=xtal.Unconventional00;	Unconventional[0][1]=xtal.Unconventional01;	Unconventional[0][2]=xtal.Unconventional02
		#		Unconventional[1][0]=xtal.Unconventional10;	Unconventional[1][1]=xtal.Unconventional11;	Unconventional[1][2]=xtal.Unconventional12
		#		Unconventional[2][0]=xtal.Unconventional20;	Unconventional[2][1]=xtal.Unconventional21;	Unconventional[2][2]=xtal.Unconventional22
		#	else
		#		KillWaves/Z root:Packages:Lattices:Unconventional
		#	endif

	def __calcDensity(self):
		""" calculate the density (g/cm^3) """
		amuAll = 0.0								# atomic mass of all atoms in cell
		for atom in self.atoms:
			amuAll += atom.amu * atom.mult * atom.occ
		self.density = (amuAll/NA)/(self.Vc * 1e-21)	# grams / cm^3
		if amuAll==0: self.density = float('nan')
		return self.density


	def Hex2Rhom(self,aH,cH):
		""" convert lattice constants
		aH,cH	Hexagonal lattice constants
		returns aR,alpha as a tuple, Rhombohedral lattice constants
		"""
		aR = math.sqrt(3.0*aH*aH + cH*cH) / 3.0
		alpha = asin( 1.5 / sqrt(3.0+(cH/aH)**2) ) * 2.0 * 180.0/math.pi
		return (aR,alpha)

	def Rhom2Hex(self,aR,alpha):
		""" convert lattice constants
		aR, alpha				Rhombohedral lattice constants
		returns tumpe(aH,cH)	Hexagonal lattice constants
		"""
		ca2 = (1.5 / sin(alpha*math.pi/90.0))**2 - 3.0	# = (cH/aH)^2
		aH  = math.sqrt( (9.0*aR*aR)/(ca2 + 3.0) )
		cH = aH * math.sqrt(ca2)
		return (aH,cH)


	def dSpacing(self,h,k,l,T=None):
		""" returns d-spacing for the hkl (nm) """
		hkl = np.matrix([h,k,l])
		qvec = self.recip * hkl.T			# qvec =  recip x hkl
		norm = np.linalg.norm(qvec)
		if norm <= 0.0: return float('inf')
		d = 2*(math.pi) / norm				# q = 2PI/d

		try:
			alphaT = float(self.alphaT)
			try: T = float(T)
			except: T = float(self.Temperature)
			d = d*(1+alphaT*(T-22.5))	# apply temperature correction
		except:
			pass
		return d

	def hkl2Q(self,h,k,l):
		""" compute qvector for (h,k,l) """
		hkl = np.matrix([h,k,l])
		qvec = self.recip * hkl.T			# qvec =  recip x hkl
		return qvec

	def uvw2ZoneDir(self,u,v,w):
		""" compute zone direction of vector for (u,v,w) """
		uvw = np.matrix([u,v,w])
		vec = self.direct * uvw.T			# vec =  direct x uvw
		return vec


	def angleBetweenHKLs(self,h1,k1,l1, h2,k2,l2):
		""" find angle between (h1,k1,l1) and (h2,k2,l2) """
		hkl1 = np.matrix([h1,k1,l1])
		hkl2 = np.matrix([h2,k2,l2])
		qvec1 = self.recip * hkl1.T			# qvec =  recip x hkl
		qvec2 = self.recip * hkl2.T
		n1 = np.linalg.norm(qvec1)
		n2 = np.linalg.norm(qvec2)
		dot = float((qvec1.T * qvec2)[0]) / (n1*n2)

		if dot>=1.0: return 0.0
		elif dot<=-1.0: return 180.0
		return math.acos(dot) * 180/math.pi


	def angleBetweenZones(self,u1,v1,w1, u2,v2,w2):
		""" find angle between two zones: (u1,v1,w1) and (u2,v2,w2) """
		uvw1 = np.matrix([u1,v1,w1])
		uvw2 = np.matrix([u2,v2,w2])
		vec1 = self.direct * uvw1.T			# vec =  direct x uvw
		vec2 = self.direct * uvw2.T
		n1 = np.linalg.norm(vec1)
		n2 = np.linalg.norm(vec2)
		dot = float((vec1.T * vec2)[0]) / (n1*n2)

		if dot>=1.0: return 0.0
		elif dot<=-1.0: return 180.0
		return math.acos(dot) * 180/math.pi


	def __ALLOW_FC(self,h,k,l):
		""" face-centered, hkl must be all even or all odd """
		return (h+k) % 2 == 0 and (k+l) % 2 == 0

	def __ALLOW_BC(sefl,h,k,l):
		""" body-centered, !mod(round(h+k+l),2), sum must be even """
		return (h+k+l) % 2 == 0

	def __ALLOW_CC(self,h,k,l):
		""" C-centered, !mod(round(h+k),2) """
		return ((h+k) % 2)==0

	def __ALLOW_AC(self,h,k,l):
		""" A-centered, !mod(round(k+l),2) """
		return ((k+l) % 2)==0

	def __ALLOW_RHOM_HEX(self,h,k,l):
		""" rhombohedral hexagonal, allowed are -H+K+L=3n or H-K+L=3n """
		return (-h+k+l)%3 == 0 or (h-k+l)%3 == 0

	def __ALLOW_HEXAGONAL(self,h,k,l):
		""" hexagonal, forbidden are: H+2K=3N with L odd """
		return bool((h+2*k)%3) or not bool(l%2)


	def Fstruct(self,h,k,l,keV=None,T_K=None):
		""" returns the structure factor,
		keV is not used yet.
		Temperature (K), used to calculate Debye-Waller factor, not used yet.
		"""
		zero = complex(0.0, 0.0)
		usingHexAxes = (abs(90-self.alpha)+abs(90-self.beta)+abs(120-self.gam))<1e-6
		system = self.__latticeSystemNum()
		TRIGONAL = 4
		HEXAGONAL = 5
		if not(h%1 or k%1 or l%1):				# non-integral, always allowed
			sym = getHMsym(self.SpaceGroup)# get symmetry symbol
			sym = sym[0]

			if sym is 'F':
				if not self.__ALLOW_FC(h,k,l): return zero
			elif sym is 'I':
				if not self.__ALLOW_BC(h,k,l): return zero
			elif sym is 'C':
				if not self.__ALLOW_CC(h,k,l): return zero
			elif sym is 'A':
				if not self.__ALLOW_AC(h,k,l): return zero
			elif sym is 'R' and usingHexAxes:
				if not self.__ALLOW_RHOM_HEX(h,k,l): return zero # rhombohedral cell with hexagonal axes
			if system==HEXAGONAL:				# Hexagonal
				if not __ALLOW_HEXAGONAL(h,k,l): return zero

		Q = 2.0 * math.pi / self.dSpacing(h,k,l)	# |Q| vector (nm)
		if self.Vibrate and self.atoms[0].U11 > 0:	# need Q-vector too.
			hkl = np.matrix([h,k,l])
			qvec = self.recip * hkl.T			# qvec =  recip x hkl
		if len(self.atoms)<1: return complex(1.0,0.0)

		Freal=0.0
		Fimag=0.0
		fatomArg = 0.0
		for atom in self.atoms:
			fatomC = atom.fatom(Q,keV)
			fatomArg = cmath.phase(fatomC)
			fatomMag = abs(fatomC) * atom.occ

			if self.Vibrate:
				# M = B*(sin^2(theta)/lam^2) --> M = B/(16 π^2) * Q^2
				# exp(-B * sin^2(theta)/lam^2)		B = 8 * π^2 * <u^2> =  8 * π^2 * Uiso
				thetaM = atom.DebyeT
				Biso = atom.Biso						# B-factor (nmÍ^2)
				Uiso = atom.Uiso

				if type(thetaM) is float and type(T_K) is float:	# have a valid Temperature and Debye Temperature
					amu = atom.amu								# mass of atom (amu)
					DW = math.exp(-atom.DW_factor_M(T_K,thetaM,Q))	# calculates the M in exp(-M), no I/O
					if DW>0: fatomMag *= DW
				elif (Biso>0):
					fatomMag *= math.exp(-Biso * (Q/(4*math.pi))^2)	# exp[-B * (sin(theta)/lam)^2 ]
				elif (Uiso>0):
					fatomMag *= math.exp(-Q*Q*Uiso/2)				# B = (8*π^2 U)
				elif atom.U11 > 0:									# qUq = q^t x U x q
					qUq = (atom.U11)*qvec[0]^2 + (atom.U22)*qvec[1]^2 + (atom.U33)*qvec[2]^2
					if type(atom.U12) is float:
						qUq += 2*(atom.U23)*qvec[1]*qvec[2] + 2*(atom.U13)*qvec[2]*qvec[0] + 2*(atom.U12)*qvec[0]*qvec[1]

					if type(qUq) is float: DW = math.exp(-qUq/2)
					else: DW = 1.0
					if DW>0.0: fatomMag *= DW

			for xyz in atom.xyz:
				(x,y,z) = xyz
				arg = 2*math.pi*(h*x + k*y + l*z) + fatomArg
				Freal += fatomMag * math.cos(arg)
				Fimag += fatomMag * math.sin(arg)

		if system==HEXAGONAL or (usingHexAxes and system==TRIGONAL):
			arg = 2*math.pi*((h+2*k)/3 + l*0.5)			# hexagonal has atoms at (0,0,0) and (1/3, 2/3, 1/2)
			Fr = 1.0 + math.cos(arg)
			Fi = math.sin(arg)
			rr = Freal
			ii = Fimag
			Freal = rr*Fr - ii*Fi
			Fimag = rr*Fi + ii*Fr
			#  for hexagonal:
			#	h+2k=3n,	l=even;		F = 4*f			1
			#	h+2k=3n±1,	l=odd;		F = sqrt(3)*f   sqrt(3)/4
			#	h+2k=3n±1,	l=even;		F = f			1/4
			#	h+2k=3n,	l=odd;		F = 0			0

		if abs(Freal)<1e-8: Freal = 0.0
		if abs(Fimag)<1e-8: Fimag = 0.0
		return complex(Freal,Fimag)


	def allowedHKL(self,h,k,l):
		Fc = self.Fstruct(h,k,l)
		if len(self.atoms)>1:	N = float(len(self.atoms))
		else:					N = 1.0
		return abs(Fc)/N > self.allowed_F_N	# allowed means more than 0.01 electron/atom


	def findClosestHKL(self,dIN):
		""" Find the hkl(s) closest to given d-spacing (nm) """
		hmax = math.ceil(self.dSpacing(1,0,0)/dIN) + 1
		kmax = math.ceil(self.dSpacing(0,1,0)/dIN) + 1
		lmax = math.ceil(self.dSpacing(0,0,1)/dIN) + 1

		saved = list()						# list of hkl closest to dIN, each element is a dict
		err0 = float('inf')
		if len(self.atoms)>1:	N = float(len(self.atoms))
		else:					N = 1.0
		for l in symrange(lmax):			# 0,1,-1,2,-2,3,-3,... lmax,-lmax
			for k in symrange(kmax):
				for h in symrange(hmax):
					Fhkl = self.Fstruct(h,k,l)
					if abs(Fhkl)/N < self.allowed_F_N: continue	# only consider allowed hkl (if too small skip it)
					dhkl = self.dSpacing(h,k,l)
					derr = abs(dhkl-dIN)
					if derr<err0:			# found a new best one
						err0 = derr
						saved = list()
					if derr<=err0:			# a duplicate best one
						saved.append({'h':h, 'k':k, 'l':l, 'dhkl':dhkl, 'Fhkl':Fhkl})

		# remove entry from saved if hkl == -hkl
		short = list()						# the new shorter list
		for d in saved:
			# check if -d is in short
			skip = False
			for ds in short:
				skip = skip or (abs(ds['h']+d['h']) + abs(ds['k']+d['k']) + abs(ds['l']+d['l'])==0)
			if not skip: short.append(d)

		return short



	def __str__(self):
		""" Return string value for Lattice. """
		SG = self.SpaceGroup

		if len(self.desc):		out = '"%s"   ' % self.desc
		out += 'Space Group=%d   %s   %s       Vc = %g (nm^3)' % (SG,self.latticeSystemName, getHMboth(self.SpaceGroup),self.Vc)
		if self.density > 0:	out += '       density = %g (g/cm^3)' % self.density

		a = self.a
		b = self.b
		c = self.c

		if not (self.alphaT is None):
			out += '     coefficient of thermal expansion is %g' % self.alphaT

		isys = self.__latticeSystemNum()
		if isys==6:					# Cubic
			out += '\nlattice constant =  %.15g nm' % a
		elif isys==5 or isys==4:	# Hexagonal or Trigonal, (using the hexagonal cell axes)
			out += '\nlattice constants  a = %.15g nm, c = %.15g nm' % (a,c)
		elif isys==3:				# Tetragonal
			out += '\nlattice constants  a = %.15gnm, c = %.15gnm' % (a,b)
		elif isys==2:				# Orthorhombic
			out += '\nlattice constants  a = %.15gnm, b = %.15gnm, c = %.15gnm' % (a,b,c)
		else:						# Monoclinic or Triclinic
			out += '\nlattice constants  { %.15gnm, %.15gnm, %.15gnm,   %.15g, %.15g, %.15g }' % (a,b,c,self.alpha,self.beta,self.gam)

		if not (self.Temperature is None):
			if self.Temperature >= -273.15:
				Temperature = self.Temperature		# self.Temperature is always C
				unit = "C"
				if Temperature < -100: unit = 'K'	# display in Kelvin
				Temperature = ConvertTemperatureUnits(Temperature,"C",unit)	# self.Temperature is always C
				out += ',   Temperature = %g %s' % (Temperature,unit)

		out += '\n'

		if (SG<=167 and SG>=143 and (abs(90-self.alpha)+abs(90-self.beta)+abs(120-self.gam))<1e-6):	# also show rhombohedral lattice constants
			aRhom = math.sqrt(3*a*a + c*c)/3
			alphaRhom = 2 * math.asin(1.5 / math.sqrt(3+(c*c/a/a))) * 180/ math.pi
			out += '   Rhombohedral constants, aRhom = %.8g nm,   alpha(Rhom) = %.8g\n' % (aRhom,alphaRhom)

		if len(self.atoms)<1:
			out += 'No Atoms Defined\n'
		else:
			if len(self.formulaMin)>0:
				out += 'atom type locations:\t chemical formula = "%s"\n' % self.formulaMin
			else:
				out += 'atom type locations:\n'
			for atom in self.atoms:
				out += '%s\n' % atom

		if len(self.bonds)>0:
			out += '\nspecified bonds:'
			for bond in self.bonds:
				out += '\n%s' % bond

		return out

	def __repr__(self):
		""" Return printable representation for a Lattice. """
		out = 'Lattice[desc=%r, SpaceGroup=%r' % (self.desc,self.SpaceGroup)
		out += ', a=%r, b=%r, c=%r, alpha=%r, beta=%r, gam=%r' % (self.a,self.b,self.c,self.alpha,self.beta,self.gam)
		out +=  ', Temperature=%r, alphaT=%r,' % (self.Temperature, self.alphaT)
		out += '\n  Vc=%r, density=%r, Vibrate=%r, haveDebyeT=%r, formula=%r,' % (self.Vc, self.density, self.Vibrate, self.haveDebyeT,self.formulaMin) 
		out += '\n  direct = %r' % self.direct
		out += '\n  recip = %r' % self.recip
		for atom in self.atoms: out += '\n%r' % atom
		for bond in self.bonds: out += '\n%r' % bond
		if len(self.atoms) or len(self.bonds): out += '\n'
		out += ']'
		return out


	def __MinimalChemFormula(self):
		""" Create a minimal chemical formula for this structure """
		if len(self.atoms)<1:
			self.formulaMin = ''
			return ''

		atomList = list()
		fdict = {}
		for atom in self.atoms:
			value = atom.occ * atom.mult
			if value <= 0: continue
			symbol = atom.sym
			try:
				prev = fdict[symbol]
			except:
				prev = 0
				atomList.append(symbol)		# atomList needed because dicts do not preserve order
			fdict[symbol] = prev + value

		nums = list()
		for item in fdict: nums.append(fdict[item])

		iMax = 0
		allInts = True
		for i in nums:
			i = int(i)
			iMax = max(i,iMax)
			if not(i == round(i)): allInts = False

		if allInts:					# remove common factors
			for factor in range(iMax,0,-1):	# check factor
				allDivide = True
				for i in nums:
					if not(i % factor == 0):
						allDivide = False
						break
				if allDivide: break
			if factor>1:			# divide out factor
				for item in fdict:
					val = fdict[item] / factor
					fdict[item] = int(val)

		# now create formula from fdict
		formula = ''
		for symbol in atomList:
			val = fdict[symbol]
			if val<=0:	continue
			formula += str(symbol)
			if val!=1:
				if val == round(val): val = int(val)
				formula += str(val)

		self.formulaMin = formula
		return formula


	def __latticeSystemNum(self):
		# returns number, 0=Triclinic,1=Monoclinic,2=Orthorhombic,3=Tetragonal,4=Trigonal,5=Hexagonal,6=Cubic
		SG = self.SpaceGroup		# Space Group number, from International Tables
		if SG > 230:	raise ValueError('Space Group is not in [1,230]')
		elif SG >= 195:	return 6	# Cubic
		elif SG >= 168:	return 5	# Hexagonal
		elif SG >= 143:	return 4	# Trigonal, (using the hexagonal cell axes)
		elif SG >= 75:	return 3	# Tetragonal
		elif SG >= 16:	return 2	# Orthorhombic
		elif SG >= 3:	return 1	# Monoclinic
		elif SG >= 0:	return 0	# Triclinic
		else:			raise ValueError('Space Group is not in [1,230]')

	def __latticeSystemName(self):
		SG = self.SpaceGroup		# Space Group number, from International Tables
		if SG > 230:	raise ValueError('Space Group is not in [1,230]')
		elif SG >= 195:	return 'Cubic'
		elif SG >= 168:	return 'Hexagonal'
		elif SG >= 143:	return 'Trigonal'
		elif SG >= 75:	return 'Tetragonal'
		elif SG >= 16:	return 'Orthorhombic'
		elif SG >= 3:	return 'Monoclinic'
		elif SG >= 0:	return 'Triclinic'
		else:			raise ValueError('Space Group is not in [1,230]')



	def SetSymmetryOperations(self):
		""" sets the symmetry operations for a SpaceGroup as an array of numpy matricies """

		if self.SpaceGroup==1:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==2:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==3:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==4:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==5:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==6:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==7:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==8:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==9:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==10:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==11:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==12:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==13:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==14:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==15:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==16:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==17:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==18:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==19:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==20:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==21:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==22:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==23:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==24:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==25:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==26:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==27:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==28:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==29:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==30:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==31:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==32:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==33:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==34:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==35:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==36:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==37:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==38:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==39:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==40:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==41:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==42:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==43:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.] ] ] )
		elif self.SpaceGroup==44:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==45:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==46:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==47:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==48:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==49:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==50:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==51:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==52:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==53:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==54:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==55:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==56:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==57:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==58:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==59:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==60:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==61:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==62:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==63:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==64:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==65:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==66:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==67:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==68:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==69:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==70:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.] ] ] )
		elif self.SpaceGroup==71:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==72:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==73:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==74:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==75:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==76:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==77:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==78:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/4.] ] ] )
		elif self.SpaceGroup==79:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==80:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==81:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==82:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==83:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==84:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==85:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==86:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==87:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==88:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 3/4.] ] ] )
		elif self.SpaceGroup==89:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==90:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==91:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/4.] ] ] )
		elif self.SpaceGroup==92:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==93:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==94:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==95:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 3/4.] ] ] )
		elif self.SpaceGroup==96:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 3/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==97:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==98:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==99:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==100:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==101:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==102:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==103:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==104:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==105:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==106:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==107:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==108:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==109:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==110:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==111:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==112:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==113:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==114:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==115:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==116:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==117:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==118:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==119:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==120:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==121:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==122:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==123:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==124:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==125:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==126:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==127:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==128:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==129:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==130:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==131:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==132:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==133:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==134:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==135:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==136:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==137:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==138:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==139:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==140:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==141:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==142:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 3/4.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 3/4.] ] ] )
		elif self.SpaceGroup==143:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==144:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ] ] )
		elif self.SpaceGroup==145:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ] ] )
		elif self.SpaceGroup==146:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ] ] )
		elif self.SpaceGroup==147:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==148:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,0.,0., 2/3.], [0.,-1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 2/3.], [-1.,1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [1.,-1.,0., 2/3.], [1.,0.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,0.,0., 1/3.], [0.,-1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 1/3.], [-1.,1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [1.,-1.,0., 1/3.], [1.,0.,0., 2/3.], [0.,0.,-1., 2/3.] ] ] )
		elif self.SpaceGroup==149:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==150:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==151:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==152:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==153:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==154:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==155:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,-1.,0., 2/3.], [0.,-1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [-1.,0.,0., 2/3.], [-1.,1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 2/3.], [1.,0.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [1.,-1.,0., 1/3.], [0.,-1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [-1.,0.,0., 1/3.], [-1.,1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 1/3.], [1.,0.,0., 2/3.], [0.,0.,-1., 2/3.] ] ] )
		elif self.SpaceGroup==156:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==157:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==158:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==159:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==160:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,0.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [1.,0.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ] ] )
		elif self.SpaceGroup==161:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 5/6.] ],
			[ [1.,0.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 5/6.] ],
			[ [0.,-1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 5/6.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 1/6.] ],
			[ [1.,0.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 1/6.] ],
			[ [0.,-1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 1/6.] ] ] )
		elif self.SpaceGroup==162:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==163:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==164:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==165:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==166:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,-1.,0., 2/3.], [0.,-1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [-1.,0.,0., 2/3.], [-1.,1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 2/3.], [1.,0.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [-1.,0.,0., 2/3.], [0.,-1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 2/3.], [-1.,1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [1.,-1.,0., 2/3.], [1.,0.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,0.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [1.,-1.,0., 1/3.], [0.,-1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [-1.,0.,0., 1/3.], [-1.,1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 1/3.], [1.,0.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [-1.,0.,0., 1/3.], [0.,-1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 1/3.], [-1.,1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [1.,-1.,0., 1/3.], [1.,0.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [1.,0.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ] ] )
		elif self.SpaceGroup==167:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 1/3.] ],
			[ [1.,-1.,0., 2/3.], [0.,-1.,0., 1/3.], [0.,0.,-1., 5/6.] ],
			[ [-1.,0.,0., 2/3.], [-1.,1.,0., 1/3.], [0.,0.,-1., 5/6.] ],
			[ [0.,1.,0., 2/3.], [1.,0.,0., 1/3.], [0.,0.,-1., 5/6.] ],
			[ [-1.,0.,0., 2/3.], [0.,-1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 2/3.], [-1.,1.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [1.,-1.,0., 2/3.], [1.,0.,0., 1/3.], [0.,0.,-1., 1/3.] ],
			[ [-1.,1.,0., 2/3.], [0.,1.,0., 1/3.], [0.,0.,1., 5/6.] ],
			[ [1.,0.,0., 2/3.], [1.,-1.,0., 1/3.], [0.,0.,1., 5/6.] ],
			[ [0.,-1.,0., 2/3.], [-1.,0.,0., 1/3.], [0.,0.,1., 5/6.] ],
			[ [1.,0.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 2/3.] ],
			[ [1.,-1.,0., 1/3.], [0.,-1.,0., 2/3.], [0.,0.,-1., 1/6.] ],
			[ [-1.,0.,0., 1/3.], [-1.,1.,0., 2/3.], [0.,0.,-1., 1/6.] ],
			[ [0.,1.,0., 1/3.], [1.,0.,0., 2/3.], [0.,0.,-1., 1/6.] ],
			[ [-1.,0.,0., 1/3.], [0.,-1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 1/3.], [-1.,1.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [1.,-1.,0., 1/3.], [1.,0.,0., 2/3.], [0.,0.,-1., 2/3.] ],
			[ [-1.,1.,0., 1/3.], [0.,1.,0., 2/3.], [0.,0.,1., 1/6.] ],
			[ [1.,0.,0., 1/3.], [1.,-1.,0., 2/3.], [0.,0.,1., 1/6.] ],
			[ [0.,-1.,0., 1/3.], [-1.,0.,0., 2/3.], [0.,0.,1., 1/6.] ] ] )
		elif self.SpaceGroup==168:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==169:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/6.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 5/6.] ] ] )
		elif self.SpaceGroup==170:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 5/6.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/6.] ] ] )
		elif self.SpaceGroup==171:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 2/3.] ] ] )
		elif self.SpaceGroup==172:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/3.] ] ] )
		elif self.SpaceGroup==173:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==174:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==175:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==176:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==177:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==178:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/6.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 5/6.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 5/6.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/6.] ] ] )
		elif self.SpaceGroup==179:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 5/6.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/6.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/6.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 5/6.] ] ] )
		elif self.SpaceGroup==180:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 2/3.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/3.] ] ] )
		elif self.SpaceGroup==181:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 2/3.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/3.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 2/3.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/3.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 2/3.] ] ] )
		elif self.SpaceGroup==182:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==183:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==184:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==185:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==186:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==187:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==188:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==189:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==190:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==191:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==192:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==193:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==194:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,1.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [1.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,-1.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [-1.,1.,0., 0], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==195:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==196:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ] ] )
		elif self.SpaceGroup==197:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==198:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==199:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ] ] )
		elif self.SpaceGroup==200:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==201:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==202:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ] ] )
		elif self.SpaceGroup==203:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/4.], [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,1., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [1.,0.,0., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.], [1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,1., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [1.,0.,0., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.], [1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,1., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [1.,0.,0., 1/4.], [0.,1.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.], [1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 3/4.], [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,1., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [1.,0.,0., 3/4.], [0.,1.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.], [1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.] ] ] )
		elif self.SpaceGroup==204:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==205:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==206:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ] ] )
		elif self.SpaceGroup==207:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==208:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==209:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==210:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.] ] ] )
		elif self.SpaceGroup==211:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==212:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.] ] ] )
		elif self.SpaceGroup==213:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.] ] ] )
		elif self.SpaceGroup==214:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.] ] ] )
		elif self.SpaceGroup==215:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==216:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==217:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==218:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==219:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==220:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 3/4.], [1.,0.,0., 3/4.] ] ] )
		elif self.SpaceGroup==221:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==222:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==223:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==224:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==225:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==226:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==227:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/4.], [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,1., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [1.,0.,0., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,1., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [1.,0.,0., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,1., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [1.,0.,0., 1/4.], [0.,1.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 3/4.], [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,1., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [1.,0.,0., 3/4.], [0.,1.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 0] ] ] )
		elif self.SpaceGroup==228:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/4.], [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,1., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [1.,0.,0., 1/4.], [0.,1.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/4.], [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,-1., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,1., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [1.,0.,0., 3/4.], [0.,1.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [0.,0.,-1., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [0.,0.,1., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 1/2.], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 1/2.], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 1/2.], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [-1.,0.,0., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,-1., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,1., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [1.,0.,0., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,-1., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [1.,0.,0., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [-1.,0.,0., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,1., 1/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 3/4.], [-1.,0.,0., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,-1., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,1., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [1.,0.,0., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.], [1.,0.,0., 3/4.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,-1.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==229:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 0], [-1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 0], [1.,0.,0., 0], [0.,0.,-1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,1., 0], [0.,-1.,0., 0] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 0], [0.,0.,-1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,-1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 0], [0.,-1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [-1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [0.,1.,0., 0], [1.,0.,0., 0], [0.,0.,1., 0] ],
			[ [1.,0.,0., 0], [0.,0.,-1., 0], [0.,-1.,0., 0] ],
			[ [1.,0.,0., 0], [0.,0.,1., 0], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 0], [0.,1.,0., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [0.,1.,0., 0], [1.,0.,0., 0] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.], [1.,0.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [-1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 1/2.], [1.,0.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,-1., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [1.,0.,0., 1/2.], [0.,0.,1., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [0.,1.,0., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [0.,1.,0., 1/2.], [1.,0.,0., 1/2.] ] ] )
		elif self.SpaceGroup==230:
			self.equivXYZ1 = np.array( [ [ [1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 0], [1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,-1.,0., 0], [0.,0.,-1., 0] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 0] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,-1., 0], [-1.,0.,0., 0], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,-1., 0], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 0] ],
			[ [0.,0.,1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [0.,1.,0., 3/4.], [1.,0.,0., 3/4.], [0.,0.,1., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,1., 3/4.], [0.,1.,0., 3/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,1., 3/4.], [0.,1.,0., 3/4.], [1.,0.,0., 3/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.], [0.,0.,1., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 0], [0.,0.,1., 1/2.] ],
			[ [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 0] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 0], [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,0.,-1., 3/4.], [0.,1.,0., 1/4.], [1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 0], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 0] ],
			[ [0.,-1.,0., 0], [0.,0.,1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 0], [0.,1.,0., 1/2.] ],
			[ [0.,0.,-1., 0], [1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 0] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,1., 1/4.], [0.,1.,0., 3/4.] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.] ],
			[ [0.,0.,1., 3/4.], [0.,-1.,0., 3/4.], [1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.] ],
			[ [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.] ],
			[ [1.,0.,0., 1/2.], [0.,1.,0., 0], [0.,0.,-1., 1/2.] ],
			[ [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.] ],
			[ [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.] ],
			[ [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.], [0.,0.,1., 0] ],
			[ [-1.,0.,0., 3/4.], [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.], [1.,0.,0., 1/4.] ],
			[ [1.,0.,0., 0], [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.] ],
			[ [0.,0.,1., 1/4.], [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.] ],
			[ [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.], [0.,-1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,-1., 1/2.], [-1.,0.,0., 1/2.] ],
			[ [0.,1.,0., 1/2.], [0.,0.,1., 0], [-1.,0.,0., 1/2.] ],
			[ [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.], [0.,1.,0., 0] ],
			[ [0.,1.,0., 0], [0.,0.,-1., 1/2.], [1.,0.,0., 1/2.] ],
			[ [0.,0.,1., 1/2.], [1.,0.,0., 0], [0.,-1.,0., 1/2.] ],
			[ [0.,0.,1., 0], [-1.,0.,0., 1/2.], [0.,1.,0., 1/2.] ],
			[ [0.,-1.,0., 1/2.], [0.,0.,1., 1/2.], [1.,0.,0., 0] ],
			[ [0.,-1.,0., 3/4.], [-1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [0.,1.,0., 1/4.], [1.,0.,0., 1/4.], [0.,0.,1., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,-1., 3/4.], [0.,-1.,0., 1/4.] ],
			[ [1.,0.,0., 1/4.], [0.,0.,1., 1/4.], [0.,1.,0., 1/4.] ],
			[ [0.,0.,-1., 1/4.], [0.,1.,0., 1/4.], [-1.,0.,0., 3/4.] ],
			[ [0.,0.,1., 1/4.], [0.,1.,0., 1/4.], [1.,0.,0., 1/4.] ] ] )
		else:
			self.equivXYZ1 = None
			raise ValueError('SpaceGroup must be an integer in range [1,230], not "%r".' % self.SpaceGroup)

		return self.equivXYZ1


	################################################################################
	################################################################################
	#		Start of read in an XML file
	def parseCIFxml(self,inputXML):
		""" take a control xml file and parse out values needed by ProcessOneDataSet()
		This is the opposite of writeInputXMLfile() """

		if not os.path.isfile(inputXML): raise ValueError('parseCIFxml(), input xml file "%r" does not exist' % inputXML)
		if os.path.getsize(inputXML)<10: raise ValueError('parseCIFxml(), input xml file "%r" is too small to be OK' % inputXML)
		a = xml.dom.minidom.parse(inputXML,parser=None)
		cif = a.childNodes[0]
		if not(cif.nodeName == u'cif'):
			raise ValueError('parseCIFxml(), cannot find <cif> node in "%s"' % inputXML)

		try:
			SpaceGroup = int(cif.getElementsByTagName('space_group_IT_number')[0].firstChild.nodeValue)
		except:
			raise ValueError('parseCIFxml(), Space Group must be integer in range [1,230]')
		if SpaceGroup<1 or SpaceGroup>230:
			raise ValueError('parseCIFxml(), Space Group must be integer in range [1,230]')

		try:	desc = cif.getElementsByTagName('chemical_name_common')[0].firstChild.nodeValue
		except:	desc = 'struct%d' % SpaceGroup

		try:
			cell = cif.getElementsByTagName('cell')[0]
		except:
			raise IndexError('parseCIFxml(), could not find "<cell>"')

		try:
			node = cif.getElementsByTagName('a')[0]
			a = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			a *= ConvertUnits2meters(unit,defaultLen=1e-10)*1e9	# want length in nm
		except:
			raise ValueError('parseCIFxml(), cannot find a')

		try:
			node = cif.getElementsByTagName('b')[0]
			b = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			b *= ConvertUnits2meters(unit,defaultLen=1e-10)*1e9	# want length in nm
		except:
			raise ValueError('parseCIFxml(), cannot find b')

		try:
			node = cif.getElementsByTagName('c')[0]
			c = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			c *= ConvertUnits2meters(unit,defaultLen=1e-10)*1e9	# want length in nm
		except:
			raise ValueError('parseCIFxml(), cannot find c')

		try:
			node = cif.getElementsByTagName('alpha')[0]
			alpha = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			alpha = ConvertAngleUnits(alpha,unit,'deg',default='deg')
		except:
			raise ValueError('parseCIFxml(), cannot find alpha')

		try:
			node = cif.getElementsByTagName('beta')[0]
			beta = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			beta = ConvertAngleUnits(beta,unit,'deg',default='deg')
		except:
			raise ValueError('parseCIFxml(), cannot find beta')

		try:
			node = cif.getElementsByTagName('gamma')[0]
			gamma = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			gamma = ConvertAngleUnits(gamma,unit,'deg',default='deg')
		except:
			raise ValueError('parseCIFxml(), cannot find gamma')

		try:
			node = cif.getElementsByTagName('temperature')[0]
			Temperature = float(node.firstChild.nodeValue)
			unit = node.getAttribute('unit')
			if len(unit): Temperature = ConvertTemperatureUnits(Tin,unit,'C')
		except:
			Temperature = None

		try:	alphaT = float(cif.getElementsByTagName('alphaT')[0].firstChild.nodeValue)
		except:	alphaT = None

		atoms = list()
		atom_sites = cif.getElementsByTagName('atom_site')
		for atom_site in atom_sites:
			try:	node = atom_site.getElementsByTagName('label')[0]
			except:	raise ValueError('parseCIFxml(), each atom site requires a label')

			label = node.firstChild.nodeValue
			if len(label)<1: raise ValueError('parseCIFxml(), each atom site requires a label')

			try:	symbol = atom_site.getElementsByTagName('symbol')[0].firstChild.nodeValue
			except: symbol = label
			Zatom = ZfromLabel(label)
			if Zatom<1:	raise ValueError('parseCIFxml(), Zatom must be in [1,'+str(Zmax)+']')

			try:	valence = int(atom_site.getElementsByTagName('valence')[0].firstChild.nodeValue)
			except: valence = 0				# valence defaults to 0 (integer)

			try:
				xyz = atom_site.getElementsByTagName('fract_xyz')[0].firstChild.nodeValue.split()
				x,y,z = xyz					# this line fails unless len(xyz)==3
				x = float(x)
				y = float(y)
				z = float(z)
				xyz = [x,y,z]				# reconstitute xyz as a float, not a list of strings
			except:
				try:
					x = float(atom_site.getElementsByTagName('fract_x')[0].firstChild.nodeValue)
					y = float(atom_site.getElementsByTagName('fract_y')[0].firstChild.nodeValue)
					z = float(atom_site.getElementsByTagName('fract_z')[0].firstChild.nodeValue)
					xyz = [x,y,z]
				except:
					x = y = z = float('nan')

			if math.isnan(x+y+z):			# cannot find valid x,y,z, give up
				raise ValueError('parseCIFxml(), cannot find fractional x,y,z')

			try:	WyckoffSymbol = atom_site.getElementsByTagName('WyckoffSymbol')[0].firstChild.nodeValue
			except: WyckoffSymbol = ''

			if (len(WyckoffSymbol) > 0) and (math.isnan(x+y+z)):	# try to set x,y,z using Wyckoff symbol
				(x,y,z) = ForceXYZtoWyckoff(SpaceGroup,WyckoffSymbol,x,y,z)

			if len(WyckoffSymbol) == 0:		# set Wyckoff symbol from x,y,z
				(WyckoffSymbol,mult) = FindWyckoffSymbol(SpaceGroup,x,y,z)
			else:
				mult = MultiplicityFromWyckoff(SpaceGroup,WyckoffSymbol)	# just find multiplicity

			try:	occ = float(atom_site.getElementsByTagName('occupancy')[0].firstChild.nodeValue)
			except: occ = 1.0

			try:
				node = atom_site.getElementsByTagName('DebyeTemperature')[0]
				DebyeT = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				if len(unit)<1: unit = 'K'
				DebyeT = ConvertTemperatureUnits(DebyeT,unit,'K')
				if DebyeT<=0: DebyeT = float('nan')
			except:
				DebyeT = float('nan')

			try:
				node = atom_site.getElementsByTagName('B_iso')[0]
				Biso = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				Biso *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if Biso<=0: Biso = float('nan')
			except:
				Biso = float('nan')

			try:
				node = atom_site.getElementsByTagName('U_iso')[0]
				Uiso = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				Uiso *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if Uiso<=0: Uiso = float('nan')
			except:
				Uiso = float('nan')

			try:
				node = atom_site.getElementsByTagName('aniso_U_11')[0]
				U11 = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				U11 *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if U11<=0: U11 = float('nan')
			except:
				U11 = float('nan')

			try:
				node = atom_site.getElementsByTagName('aniso_U_22')[0]
				U22 = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				U22 *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if U22<=0: U22 = float('nan')
			except:
				U22 = float('nan')

			try:
				node = atom_site.getElementsByTagName('aniso_U_33')[0]
				U33 = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				U33 *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if U33<=0: U33 = float('nan')
			except:
				U33 = float('nan')

			try:
				node = atom_site.getElementsByTagName('aniso_U_12')[0]
				U12 = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				U12 *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if U12<=0: U12 = float('nan')
			except:
				U12 = float('nan')

			try:
				node = atom_site.getElementsByTagName('aniso_U_13')[0]
				U13 = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				U13 *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if U13<=0: U13 = float('nan')
			except:
				U13 = float('nan')

			try:
				node = atom_site.getElementsByTagName('aniso_U_23')[0]
				U23 = float(node.firstChild.nodeValue)
				unit = node.getAttribute('unit')
				U23 *= ConvertUnits2meters(unit,defaultLen=1e-20)*1e18	# want length in nm^2
				if U23<=0: U23 = float('nan')
			except:
				U23 = float('nan')

			atom = atomXtal(label, xyz[0],xyz[1],xyz[2], Zatom, valence, occ, WyckoffSymbol,None,mult, DebyeT,Biso,Uiso,U11,U22,U33,U12,U13,U23)
			atoms.append(atom)

		bonds = list()
		bondNodes = cif.getElementsByTagName('bond_chemical')
		for node in bondNodes:
			unit = node.getAttribute('unit')
			n0 = node.getAttribute('n0')
			n1 = node.getAttribute('n1')

			sss = node.firstChild.nodeValue
			try:	lengths = [float(sss)]
			except:	lengths = node.firstChild.nodeValue.split()

			if len(n0)<1 or len(n1)<1 or len(lengths)<1:
				raise ValueError('parseCIFxml(), invalid bond, n0="%s", n1="%s", lengths = "%r"' % (n0,n1,node.firstChild.nodeValue))
			factor = ConvertUnits2meters(unit,defaultLen=1e-10)*1e9 # want length in nm
			if not (factor==1.0):
				for i in range(len(lengths)):
					lengths[i] *= factor

			bond = bondType(n0,n1,lengths)
			bonds.append(bond)

		# set the values
		self.__init__(SpaceGroup, a,b,c,alpha,beta,gamma, desc=desc, Temperature=Temperature,alphaT=alphaT,atoms=atoms,bonds=bonds)
	#		End of read in an XML file
	################################################################################



"""
some useful utility type routines:
separate utilities:
getHMboth(SpaceGroup)
getHMsym(SpaceGroup)
getFullHMSym(SpaceGroup)
getHallSymbol(SpaceGroup)
isRhombohedral(SG)				#True for those Space Groups that can be Rhombohedral
SymString2SGs(symFind,type)		# returns list of possible space groups
								# symFind, requested symbol, with possible '*'
								# type, 0=all, 1=Hermann-Mauguin, 2=Full Hermann-Mauguin, 3=Hall, 4=Lattice System
zeroOut(value)					# sets values close enough to zero to really zero
"""



def SymString2SGs(symFind,type):
	""" finds space group of a Hermann-Mauguin or Hall symbol, wild '*' cards allowed
	symFind			requested symbol, if empty, then a dialog will come up
	type			0=all, 1=Hermann-Mauguin, 2=Full Hermann-Mauguin, 3=Hall, 4=Lattice System
	returns a list of Space Group numbers that match
	"""
	type = int(type)
	symFind = symFind.replace(' ','')	# do not include spaces in search
	symFind = symFind.lower()			# search is case insensitive

	tlist = list()
	if type == 1 or type == 0:			# Hermann-Mauguin symbols
		for SG in range(1,231):			# check all 230 space goups [1,230]
			sym = getHMsym(SG)
			sym = sym.replace(' ','')	# do not include spaces in search
			sym = sym.lower()			# search is case insensitive
			if fnmatch(sym,symFind): tlist.append(SG)

	if type == 2 or type == 0:			# full Hermann-Mauguin symbols
		for SG in range(1,231):			# check all 230 space goups [1,230]
			sym = getFullHMSym(SG)
			sym = sym.replace(' ','')	# do not include spaces in search
			sym = sym.lower()			# search is case insensitive
			if fnmatch(sym,symFind): tlist.append(SG)

	if type == 3 or type == 0:			# Hall symbols
		for SG in range(1,231):			# check all 230 space goups [1,230]
			sym = getHallSymbol(SG)
			sym = sym.replace(' ','')	# do not include spaces in search
			sym = sym.lower()			# search is case insensitive
			if fnmatch(sym,symFind): tlist.append(SG)

	if type == 4 or type == 0:			# Lattice Systems
		if fnmatch('triclinic',symFind): tlist += [1,2]				# adds [1-2]
		if fnmatch('monoclinic',symFind): tlist += range(3,16)		# adds [3-15]
		if fnmatch('orthorhombic',symFind): tlist += range(16,75)	# adds [16-74]
		if fnmatch('tetragonal',symFind): tlist += range(75,143)	# adds [75-142]
		if fnmatch('trigonal',symFind): tlist += range(143,168)		# adds [143-167]
		if fnmatch('rhombohedral',symFind): tlist += range(143,168)	# adds [143-167]
		if fnmatch('hexagonal',symFind): tlist += range(168,194)	# adds [168-193]
		if fnmatch('cubic',symFind): tlist += range(194,231)		# adds [195-230]

	tlist = list(set(tlist))			# remove possible duplicates
	return tlist



def getHMboth(SG):
	""" returns short and (full) Hermann-Mauguin symbol """
	short = getHMsym(SG)
	full = getFullHMSym(SG)
	if short == full:	return short
	else:				return short + '  ('+full+')'


def getHMsym(SG):
	""" returns short Hermann-Mauguin symbol """
	# there are 230 items in this list
	symms= ('P1','P-1','P2:b','P21:b','C2:b1','Pm:b','Pc:b1','Cm:b1','Cc:b1','P2/m:b',
	'P21/m:b','C2/m:b1','P2/c:b1','P21/c:b1','C2/c:b1','P222','P2221','P21212','P212121',
	'C2221','C222','F222','I222','I212121','Pmm2','Pmc21','Pcc2','Pma2','Pca21','Pnc2',
	'Pmn21','Pba2','Pna21','Pnn2','Cmm2','Cmc21','Ccc2','Amm2','Abm2','Ama2','Aba2',
	'Fmm2','Fdd2','Imm2','Iba2','Ima2','Pmmm','Pnnn:1','Pccm','Pban:1','Pmma','Pnna',
	'Pmna','Pcca','Pbam','Pccn','Pbcm','Pnnm','Pmmn:1','Pbcn','Pbca','Pnma','Cmcm',
	'Cmca','Cmmm','Cccm','Cmma','Ccca:1','Fmmm','Fddd:1','Immm','Ibam','Ibca','Imma',
	'P4','P41','P42','P43','I4','I41','P-4','I-4','P4/m','P42/m','P4/n:1','P42/n:1',
	'I4/m','I41/a:1','P422','P4212','P4122','P41212','P4222','P42212','P4322','P43212',
	'I422','I4122','P4mm','P4bm','P42cm','P42nm','P4cc','P4nc','P42mc','P42bc','I4mm',
	'I4cm','I41md','I41cd','P-42m','P-42c','P-421m','P-421c','P-4m2','P-4c2','P-4b2',
	'P-4n2','I-4m2','I-4c2','I-42m','I-42d','P4/mmm','P4/mcc','P4/nbm:1','P4/nnc:1',
	'P4/mbm','P4/mnc','P4/nmm:1','P4/ncc:1','P42/mmc','P42/mcm','P42/nbc:1','P42/nnm:1',
	'P42/mbc','P42/mnm','P42/nmc:1','P42/ncm:1','I4/mmm','I4/mcm','I41/amd:1',
	'I41/acd:1','P3','P31','P32','R3:H','P-3','R-3:H','P312','P321','P3112','P3121',
	'P3212','P3221','R32:H','P3m1','P31m','P3c1','P31c','R3m:H','R3c:H','P-31m',
	'P-31c','P-3m1','P-3c1','R-3m:H','R-3c:H','P6','P61','P65','P62','P64','P63','P-6',
	'P6/m','P63/m','P622','P6122','P6522','P6222','P6422','P6322','P6mm','P6cc',
	'P63cm','P63mc','P-6m2','P-6c2','P-62m','P-62c','P6/mmm','P6/mcc','P63/mcm',
	'P63/mmc','P23','F23','I23','P213','I213','Pm-3','Pn-3:1','Fm-3','Fd-3:1','Im-3',
	'Pa-3','Ia-3','P432','P4232','F432','F4132','I432','P4332','P4132','I4132','P-43m',
	'F-43m','I-43m','P-43n','F-43c','I-43d','Pm-3m','Pn-3n:1','Pm-3n','Pn-3m:1',
	'Fm-3m','Fm-3c','Fd-3m:1','Fd-3c:1','Im-3m','Ia-3d')
	try:	return symms[ SG -1 ]
	except:	raise ValueError('Space Group is not in [1,230]')


def getFullHMSym(SG):
	""" returns full Hermann-Mauguin symbol
	Full H-M symbols only differ from the regular ones (in getHMsym)
	for Space Groups: 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
	"""
	fullSymms=('P1','P-1','P121:b','P1211:b','C121:b1','P1m1:b','P1c1:b1','C1m1:b1',
	'C1c1:b1','P12/m1:b','P121/m1:b','C12/m1:b1','P12/c1:b1','P121/c1:b1','C12/c1:b1',
	'P222','P2221','P21212','P212121','C2221','C222','F222','I222','I212121','Pmm2',
	'Pmc21','Pcc2','Pma2','Pca21','Pnc2','Pmn21','Pba2','Pna21','Pnn2','Cmm2','Cmc21',
	'Ccc2','Amm2','Abm2','Ama2','Aba2','Fmm2','Fdd2','Imm2','Iba2','Ima2','Pmmm',
	'Pnnn:1','Pccm','Pban:1','Pmma','Pnna','Pmna','Pcca','Pbam','Pccn','Pbcm','Pnnm',
	'Pmmn:1','Pbcn','Pbca','Pnma','Cmcm','Cmca','Cmmm','Cccm','Cmma','Ccca:1','Fmmm',
	'Fddd:1','Immm','Ibam','Ibca','Imma','P4','P41','P42','P43','I4','I41','P-4','I-4',
	'P4/m','P42/m','P4/n:1','P42/n:1','I4/m','I41/a:1','P422','P4212','P4122','P41212',
	'P4222','P42212','P4322','P43212','I422','I4122','P4mm','P4bm','P42cm','P42nm',
	'P4cc','P4nc','P42mc','P42bc','I4mm','I4cm','I41md','I41cd','P-42m','P-42c','P-421m',
	'P-421c','P-4m2','P-4c2','P-4b2','P-4n2','I-4m2','I-4c2','I-42m','I-42d','P4/mmm',
	'P4/mcc','P4/nbm:1','P4/nnc:1','P4/mbm','P4/mnc','P4/nmm:1','P4/ncc:1','P42/mmc',
	'P42/mcm','P42/nbc:1','P42/nnm:1','P42/mbc','P42/mnm','P42/nmc:1','P42/ncm:1',
	'I4/mmm','I4/mcm','I41/amd:1','I41/acd:1','P3','P31','P32','R3:H','P-3','R-3:H',
	'P312','P321','P3112','P3121','P3212','P3221','R32:H','P3m1','P31m','P3c1','P31c',
	'R3m:H','R3c:H','P-31m','P-31c','P-3m1','P-3c1','R-3m:H','R-3c:H','P6','P61','P65',
	'P62','P64','P63','P-6','P6/m','P63/m','P622','P6122','P6522','P6222','P6422',
	'P6322','P6mm','P6cc','P63cm','P63mc','P-6m2','P-6c2','P-62m','P-62c','P6/mmm',
	'P6/mcc','P63/mcm','P63/mmc','P23','F23','I23','P213','I213','Pm-3','Pn-3:1','Fm-3',
	'Fd-3:1','Im-3','Pa-3','Ia-3','P432','P4232','F432','F4132','I432','P4332','P4132',
	'I4132','P-43m','F-43m','I-43m','P-43n','F-43c','I-43d','Pm-3m','Pn-3n:1','Pm-3n',
	'Pn-3m:1','Fm-3m','Fm-3c','Fd-3m:1','Fd-3c:1','Im-3m','Ia-3d')
	try:	return fullSymms[ SG -1 ]
	except:	raise ValueError('Space Group is not in [1,230]')


def getHallSymbol(SG):
	""" Hall Symbols """
	Hall = ('P 1','-P 1','P 2y:b','P 2yb:b','C 2y:b1','P -2y:b','P -2yc:b1','C -2y:b1',
	'C -2yc:b1','-P 2y:b','-P 2yb:b','-C 2y:b1','-P 2yc:b1','-P 2ybc:b1','-C 2yc:b1',
	'P 2 2','P 2c 2','P 2 2ab','P 2ac 2ab','C 2c 2','C 2 2','F 2 2','I 2 2','I 2b 2c',
	'P 2 -2','P 2c -2','P 2 -2c','P 2 -2a','P 2c -2ac','P 2 -2bc','P 2ac -2','P 2 -2ab',
	'P 2c -2n','P 2 -2n','C 2 -2','C 2c -2','C 2 -2c','A 2 -2','A 2 -2c','A 2 -2a',
	'A 2 -2ac','F 2 -2','F 2 -2d','I 2 -2','I 2 -2c','I 2 -2a','-P 2 2','P 2 2 -1n:1',
	'-P 2 2c','P 2 2 -1ab:1','-P 2a 2a','-P 2a 2bc','-P 2ac 2','-P 2a 2ac','-P 2 2ab',
	'-P 2ab 2ac','-P 2c 2b','-P 2 2n','P 2 2ab -1ab:1','-P 2n 2ab','-P 2ac 2ab',
	'-P 2ac 2n','-C 2c 2','-C 2bc 2','-C 2 2','-C 2 2c','-C 2b 2','C 2 2 -1bc:1',
	'-F 2 2','F 2 2 -1d:1','-I 2 2','-I 2 2c','-I 2b 2c','-I 2b 2','P 4','P 4w','P 4c',
	'P 4cw','I 4','I 4bw','P -4','I -4','-P 4','-P 4c','P 4ab -1ab:1','P 4n -1n:1',
	'-I 4','I 4bw -1bw:1','P 4 2','P 4ab 2ab','P 4w 2c','P 4abw 2nw','P 4c 2',
	'P 4n 2n','P 4cw 2c','P 4nw 2abw','I 4 2','I 4bw 2bw','P 4 -2','P 4 -2ab','P 4c -2c',
	'P 4n -2n','P 4 -2c','P 4 -2n','P 4c -2','P 4c -2ab','I 4 -2','I 4 -2c','I 4bw -2',
	'I 4bw -2c','P -4 2','P -4 2c','P -4 2ab','P -4 2n','P -4 -2','P -4 -2c',
	'P -4 -2ab','P -4 -2n','I -4 -2','I -4 -2c','I -4 2','I -4 2bw','-P 4 2','-P 4 2c',
	'P 4 2 -1ab:1','P 4 2 -1n:1','-P 4 2ab','-P 4 2n','P 4ab 2ab -1ab:1',
	'P 4ab 2n -1ab:1','-P 4c 2','-P 4c 2c','P 4n 2c -1n:1','P 4n 2 -1n:1','-P 4c 2ab',
	'-P 4n 2n','P 4n 2n -1n:1','P 4n 2ab -1n:1','-I 4 2','-I 4 2c','I 4bw 2bw -1bw:1',
	'I 4bw 2aw -1bw:1','P 3','P 31','P 32','R 3:H','-P 3','-R 3:H','P 3 2','P 3 2"',
	'P 31 2c (0 0 1)','P 31 2"','P 32 2c (0 0 -1)','P 32 2"','R 3 2":H','P 3 -2"',
	'P 3 -2','P 3 -2"c','P 3 -2c','R 3 -2":H','R 3 -2"c:H','-P 3 2','-P 3 2c',
	'-P 3 2"','-P 3 2"c','-R 3 2":H','-R 3 2"c:H','P 6','P 61','P 65','P 62','P 64',
	'P 6c','P -6','-P 6','-P 6c','P 6 2','P 61 2 (0 0 -1)','P 65 2 (0 0 1)',
	'P 62 2c (0 0 1)','P 64 2c (0 0 -1)','P 6c 2c','P 6 -2','P 6 -2c','P 6c -2',
	'P 6c -2c','P -6 2','P -6c 2','P -6 -2','P -6c -2c','-P 6 2','-P 6 2c','-P 6c 2',
	'-P 6c 2c','P 2 2 3','F 2 2 3','I 2 2 3','P 2ac 2ab 3','I 2b 2c 3','-P 2 2 3',
	'P 2 2 3 -1n:1','-F 2 2 3','F 2 2 3 -1d:1','-I 2 2 3','-P 2ac 2ab 3','-I 2b 2c 3',
	'P 4 2 3','P 4n 2 3','F 4 2 3','F 4d 2 3','I 4 2 3','P 4acd 2ab 3','P 4bd 2ab 3',
	'I 4bd 2c 3','P -4 2 3','F -4 2 3','I -4 2 3','P -4n 2 3','F -4c 2 3',
	'I -4bd 2c 3','-P 4 2 3','P 4 2 3 -1n:1','-P 4n 2 3','P 4n 2 3 -1n:1','-F 4 2 3',
	'-F 4c 2 3','F 4d 2 3 -1d:1','F 4d 2 3 -1cd:1','-I 4 2 3','-I 4bd 2c 3')
	try:	return Hall[ SG -1 ]
	except:	raise ValueError('Space Group is not in [1,230]')


def isRhombohedral(SG):
	""" True for those Space Groups that can be Rhombohedral """
	r = set([146,148,155,160,161,166,167])
	try:
		return SG in r
	except:
		raise ValueError('Space Group is not an integer')



def FindWyckoffSymbol(SG,x,y,z):
	WyckList = GetWyckoffSymList(SG)
	xyz = (x,y,z)
	for item in WyckList:
		OK = True
		for op,val in zip(item[1],xyz):
			op = op.replace('x',str(float(x)))
			op = op.replace('y',str(float(y)))
			op = op.replace('z',str(float(z)))
			try:	OK = OK and (abs(val-eval(op)) < 0.0001)
			except:	OK = False

		if OK: return (item[0],item[2])
	return ('',0)


def ForceXYZtoWyckoff(SG,symbol,x0,y0,z0):
	WyckList = GetWyckoffSymList(SG)
	for item in WyckList:
		if item[0] == symbol: break
	if not(item[0] == symbol): return None

	out = tuple()
	for op in item[1]:
		op = op.replace('x',str(float(x0)))
		op = op.replace('y',str(float(y0)))
		op = op.replace('z',str(float(z0)))
		out += (eval(op),)

	return out


def MultiplicityFromWyckoff(SG,symbol):
	""" returns multiplicity """
	WyckList = GetWyckoffSymList(SG)
	for item in WyckList:
		if item[0] == symbol: return int(item[2])
	return 0





def GetWyckoffSymList(SG):
	if SG==1: WyckList = [('a',('x','y','z'),1)]
	elif SG==2: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0','0.5','0'),1),('d',('0.5','0','0'),1),('e',('0.5','0.5','0'),1),('f',('0.5','0','0.5'),1),('g',('0','0.5','0.5'),1),('h',('0.5','0.5','0.5'),1),('i',('x','y','z'),2)]
	elif SG==3: WyckList = [('a',('0','y','0'),1),('b',('0','y','0.5'),1),('c',('0.5','y','0'),1),('d',('0.5','y','0.5'),1),('e',('x','y','z'),2)]
	elif SG==4: WyckList = [('a',('x','y','z'),2)]
	elif SG==5: WyckList = [('a',('0','y','0'),2),('b',('0','y','0.5'),2),('c',('x','y','z'),4)]
	elif SG==6: WyckList = [('a',('x','0','z'),1),('b',('x','0.5','z'),1),('c',('x','y','z'),2)]
	elif SG==7: WyckList = [('a',('x','y','z'),2)]
	elif SG==8: WyckList = [('a',('x','0','z'),2),('b',('x','y','z'),4)]
	elif SG==9: WyckList = [('a',('x','y','z'),4)]
	elif SG==10: WyckList = [('a',('0','0','0'),1),('b',('0','0.5','0'),1),('c',('0','0','0.5'),1),('d',('0.5','0','0'),1),('e',('0.5','0.5','0'),1),('f',('0','0.5','0.5'),1),('g',('0.5','0','0.5'),1),('h',('0.5','0.5','0.5'),1),('i',('0','y','0'),2),('j',('0.5','y','0'),2),('k',('0','y','0.5'),2),('l',('0.5','y','0.5'),2),('m',('x','0','z'),2),('n',('x','0.5','z'),2),('o',('x','y','z'),4)]
	elif SG==11: WyckList = [('a',('0','0','0'),2),('b',('0.5','0','0'),2),('c',('0','0','0.5'),2),('d',('0.5','0','0.5'),2),('e',('x','0.25','z'),2),('f',('x','y','z'),4)]
	elif SG==12: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0'),2),('c',('0','0','0.5'),2),('d',('0','0.5','0.5'),2),('e',('0.25','0.25','0'),4),('f',('0.25','0.25','0.5'),4),('g',('0','y','0'),4),('h',('0','y','0.5'),4),('i',('x','0','z'),4),('j',('x','y','z'),8)]
	elif SG==13: WyckList = [('a',('0','0','0'),2),('b',('0.5','0.5','0'),2),('c',('0','0.5','0'),2),('d',('0.5','0','0'),2),('e',('0','y','0.25'),2),('f',('0.5','y','0.25'),2),('g',('x','y','z'),4)]
	elif SG==14: WyckList = [('a',('0','0','0'),2),('b',('0.5','0','0'),2),('c',('0','0','0.5'),2),('d',('0.5','0','0.5'),2),('e',('x','y','z'),4)]
	elif SG==15: WyckList = [('a',('0','0','0'),4),('b',('0','0.5','0'),4),('c',('0.25','0.25','0'),4),('d',('0.25','0.25','0.5'),4),('e',('0','y','0.25'),4),('f',('x','y','z'),8)]
	elif SG==16: WyckList = [('a',('0','0','0'),1),('b',('0.5','0','0'),1),('c',('0','0.5','0'),1),('d',('0','0','0.5'),1),('e',('0.5','0.5','0'),1),('f',('0.5','0','0.5'),1),('g',('0','0.5','0.5'),1),('h',('0.5','0.5','0.5'),1),('i',('x','0','0'),2),('j',('x','0','0.5'),2),('k',('x','0.5','0'),2),('l',('x','0.5','0.5'),2),('m',('0','y','0'),2),('n',('0','y','0.5'),2),('o',('0.5','y','0'),2),('p',('0.5','y','0.5'),2),('q',('0','0','z'),2),('r',('0.5','0','z'),2),('s',('0','0.5','z'),2),('t',('0.5','0.5','z'),2),('u',('x','y','z'),4)]
	elif SG==17: WyckList = [('a',('x','0','0'),2),('b',('x','0.5','0'),2),('c',('0','y','0.25'),2),('d',('0.5','y','0.25'),2),('e',('x','y','z'),4)]
	elif SG==18: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('x','y','z'),4)]
	elif SG==19: WyckList = [('a',('x','y','z'),4)]
	elif SG==20: WyckList = [('a',('x','0','0'),4),('b',('0','y','0.25'),4),('c',('x','y','z'),8)]
	elif SG==21: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0'),2),('c',('0.5','0','0.5'),2),('d',('0','0','0.5'),2),('e',('x','0','0'),4),('f',('x','0','0.5'),4),('g',('0','y','0'),4),('h',('0','y','0.5'),4),('i',('0','0','z'),4),('j',('0','0.5','z'),4),('k',('0.25','0.25','z'),4),('l',('x','y','z'),8)]
	elif SG==22: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0.25','0.25','0.25'),4),('d',('0.25','0.25','0.75'),4),('e',('x','0','0'),8),('f',('0','y','0'),8),('g',('0','0','z'),8),('h',('0.25','0.25','z'),8),('i',('0.25','y','0.25'),8),('j',('x','0.25','0.25'),8),('k',('x','y','z'),16)]
	elif SG==23: WyckList = [('a',('0','0','0'),2),('b',('0.5','0','0'),2),('c',('0','0','0.5'),2),('d',('0','0.5','0'),2),('e',('x','0','0'),4),('f',('x','0','0.5'),4),('g',('0','y','0'),4),('h',('0.5','y','0'),4),('i',('0','0','z'),4),('j',('0','0.5','z'),4),('k',('x','y','z'),8)]
	elif SG==24: WyckList = [('a',('x','0','0.25'),4),('b',('0.25','y','0'),4),('c',('0','0.25','z'),4),('d',('x','y','z'),8)]
	elif SG==25: WyckList = [('a',('0','0','z'),1),('b',('0','0.5','z'),1),('c',('0.5','0','z'),1),('d',('0.5','0.5','z'),1),('e',('x','0','z'),2),('f',('x','0.5','z'),2),('g',('0','y','z'),2),('h',('0.5','y','z'),2),('i',('x','y','z'),4)]
	elif SG==26: WyckList = [('a',('0','y','z'),2),('b',('0.5','y','z'),2),('c',('x','y','z'),4)]
	elif SG==27: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('0.5','0','z'),2),('d',('0.5','0.5','z'),2),('e',('x','y','z'),4)]
	elif SG==28: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('0.25','y','z'),2),('d',('x','y','z'),4)]
	elif SG==29: WyckList = [('a',('x','y','z'),4)]
	elif SG==30: WyckList = [('a',('0','0','z'),2),('b',('0.5','0','z'),2),('c',('x','y','z'),4)]
	elif SG==31: WyckList = [('a',('0','y','z'),2),('b',('x','y','z'),4)]
	elif SG==32: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('x','y','z'),4)]
	elif SG==33: WyckList = [('a',('x','y','z'),4)]
	elif SG==34: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('x','y','z'),4)]
	elif SG==35: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('0.25','0.25','z'),4),('d',('x','0','z'),4),('e',('0','y','z'),4),('f',('x','y','z'),8)]
	elif SG==36: WyckList = [('a',('0','y','z'),4),('b',('x','y','z'),8)]
	elif SG==37: WyckList = [('a',('0','0','z'),4),('b',('0','0.5','z'),4),('c',('0.25','0.25','z'),4),('d',('x','y','z'),8)]
	elif SG==38: WyckList = [('a',('0','0','z'),2),('b',('0.5','0','z'),2),('c',('x','0','z'),4),('d',('0','y','z'),4),('e',('0.5','y','z'),4),('f',('x','y','z'),8)]
	elif SG==39: WyckList = [('a',('0','0','z'),4),('b',('0.5','0','z'),4),('c',('x','0.25','z'),4),('d',('x','y','z'),8)]
	elif SG==40: WyckList = [('a',('0','0','z'),4),('b',('0.25','y','z'),4),('c',('x','y','z'),8)]
	elif SG==41: WyckList = [('a',('0','0','z'),4),('b',('x','y','z'),8)]
	elif SG==42: WyckList = [('a',('0','0','z'),4),('b',('0.25','0.25','z'),8),('c',('0','y','z'),8),('d',('x','0','z'),8),('e',('x','y','z'),16)]
	elif SG==43: WyckList = [('a',('0','0','z'),8),('b',('x','y','z'),16)]
	elif SG==44: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),2),('c',('x','0','z'),4),('d',('0','y','z'),4),('e',('x','y','z'),8)]
	elif SG==45: WyckList = [('a',('0','0','z'),4),('b',('0','0.5','z'),4),('c',('x','y','z'),8)]
	elif SG==46: WyckList = [('a',('0','0','z'),4),('b',('0.25','y','z'),4),('c',('x','y','z'),8)]
	elif SG==47: WyckList = [('a',('0','0','0'),1),('b',('0.5','0','0'),1),('c',('0','0','0.5'),1),('d',('0.5','0','0.5'),1),('e',('0','0.5','0'),1),('f',('0.5','0.5','0'),1),('g',('0','0.5','0.5'),1),('h',('0.5','0.5','0.5'),1),('i',('x','0','0'),2),('j',('x','0','0.5'),2),('k',('x','0.5','0'),2),('l',('x','0.5','0.5'),2),('m',('0','y','0'),2),('n',('0','y','0.5'),2),('o',('0.5','y','0'),2),('p',('0.5','y','0.5'),2),('q',('0','0','z'),2),('r',('0','0.5','z'),2),('s',('0.5','0','z'),2),('t',('0.5','0.5','z'),2),('u',('0','y','z'),4),('v',('0.5','y','z'),4),('w',('x','0','z'),4),('x',('x','0.5','z'),4),('y',('x','y','0'),4),('z',('x','y','0.5'),4),('A',('x','y','z'),8)]
	elif SG==48: WyckList = [('a',('0.25','0.25','0.25'),2),('b',('0.75','0.25','0.25'),2),('c',('0.25','0.25','0.75'),2),('d',('0.25','0.75','0.25'),2),('e',('0.5','0.5','0.5'),4),('f',('0','0','0'),4),('g',('x','0.25','0.25'),4),('h',('x','0.25','0.75'),4),('i',('0.25','y','0.25'),4),('j',('0.75','y','0.25'),4),('k',('0.25','0.25','z'),4),('l',('0.25','0.75','z'),4),('m',('x','y','z'),8)]
	elif SG==49: WyckList = [('a',('0','0','0'),2),('b',('0.5','0.5','0'),2),('c',('0','0.5','0'),2),('d',('0.5','0','0'),2),('e',('0','0','0.25'),2),('f',('0.5','0','0.25'),2),('g',('0','0.5','0.25'),2),('h',('0.5','0.5','0.25'),2),('i',('x','0','0.25'),4),('j',('x','0.5','0.25'),4),('k',('0','y','0.25'),4),('l',('0.5','y','0.25'),4),('m',('0','0','z'),4),('n',('0.5','0.5','z'),4),('o',('0','0.5','z'),4),('p',('0.5','0','z'),4),('q',('x','y','0'),4),('r',('x','y','z'),8)]
	elif SG==50: WyckList = [('a',('0.25','0.25','0'),2),('b',('0.75','0.25','0'),2),('c',('0.75','0.25','0.5'),2),('d',('0.25','0.25','0.5'),2),('e',('0','0','0'),4),('f',('0','0','0.5'),4),('g',('x','0.25','0'),4),('h',('x','0.25','0.5'),4),('i',('0.25','y','0'),4),('j',('0.25','y','0.5'),4),('k',('0.25','0.25','z'),4),('l',('0.25','0.75','z'),4),('m',('x','y','z'),8)]
	elif SG==51: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0'),2),('c',('0','0','0.5'),2),('d',('0','0.5','0.5'),2),('e',('0.25','0','z'),2),('f',('0.25','0.5','z'),2),('g',('0','y','0'),4),('h',('0','y','0.5'),4),('i',('x','0','z'),4),('j',('x','0.5','z'),4),('k',('0.25','y','z'),4),('l',('x','y','z'),8)]
	elif SG==52: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0.25','0','z'),4),('d',('x','0.25','0.25'),4),('e',('x','y','z'),8)]
	elif SG==53: WyckList = [('a',('0','0','0'),2),('b',('0.5','0','0'),2),('c',('0.5','0.5','0'),2),('d',('0','0.5','0'),2),('e',('x','0','0'),4),('f',('x','0.5','0'),4),('g',('0.25','y','0.25'),4),('h',('0','y','z'),4),('i',('x','y','z'),8)]
	elif SG==54: WyckList = [('a',('0','0','0'),4),('b',('0','0.5','0'),4),('c',('0','y','0.25'),4),('d',('0.25','0','z'),4),('e',('0.25','0.5','z'),4),('f',('x','y','z'),8)]
	elif SG==55: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),2),('d',('0','0.5','0.5'),2),('e',('0','0','z'),4),('f',('0','0.5','z'),4),('g',('x','y','0'),4),('h',('x','y','0.5'),4),('i',('x','y','z'),8)]
	elif SG==56: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0.25','0.25','z'),4),('d',('0.25','0.75','z'),4),('e',('x','y','z'),8)]
	elif SG==57: WyckList = [('a',('0','0','0'),4),('b',('0.5','0','0'),4),('c',('x','0.25','0'),4),('d',('x','y','0.25'),4),('e',('x','y','z'),8)]
	elif SG==58: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),2),('d',('0','0.5','0.5'),2),('e',('0','0','z'),4),('f',('0','0.5','z'),4),('g',('x','y','0'),4),('h',('x','y','z'),8)]
	elif SG==59: WyckList = [('a',('0.25','0.25','z'),2),('b',('0.25','0.75','z'),2),('c',('0','0','0'),4),('d',('0','0','0.5'),4),('e',('0.25','y','z'),4),('f',('x','0.25','z'),4),('g',('x','y','z'),8)]
	elif SG==60: WyckList = [('a',('0','0','0'),4),('b',('0','0.5','0'),4),('c',('0','y','0.25'),4),('d',('x','y','z'),8)]
	elif SG==61: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('x','y','z'),8)]
	elif SG==62: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('x','0.25','z'),4),('d',('x','y','z'),8)]
	elif SG==63: WyckList = [('a',('0','0','0'),4),('b',('0','0.5','0'),4),('c',('0','y','0.25'),4),('d',('0.25','0.25','0'),8),('e',('x','0','0'),8),('f',('0','y','z'),8),('g',('x','y','0.25'),8),('h',('x','y','z'),16)]
	elif SG==64: WyckList = [('a',('0','0','0'),4),('b',('0.5','0','0'),4),('c',('0.25','0.25','0'),8),('d',('x','0','0'),8),('e',('0.25','y','0.25'),8),('f',('0','y','z'),8),('g',('x','y','z'),16)]
	elif SG==65: WyckList = [('a',('0','0','0'),2),('b',('0.5','0','0'),2),('c',('0.5','0','0.5'),2),('d',('0','0','0.5'),2),('e',('0.25','0.25','0'),4),('f',('0.25','0.25','0.5'),4),('g',('x','0','0'),4),('h',('x','0','0.5'),4),('i',('0','y','0'),4),('j',('0','y','0.5'),4),('k',('0','0','z'),4),('l',('0','0.5','z'),4),('m',('0.25','0.25','z'),8),('n',('0','y','z'),8),('o',('x','0','z'),8),('p',('x','y','0'),8),('q',('x','y','0.5'),8),('r',('x','y','z'),16)]
	elif SG==66: WyckList = [('a',('0','0','0.25'),4),('b',('0','0.5','0.25'),4),('c',('0','0','0'),4),('d',('0','0.5','0'),4),('e',('0.25','0.25','0'),4),('f',('0.25','0.75','0'),4),('g',('x','0','0.25'),8),('h',('0','y','0.25'),8),('i',('0','0','z'),8),('j',('0','0.5','z'),8),('k',('0.25','0.25','z'),8),('l',('x','y','0'),8),('m',('x','y','z'),16)]
	elif SG==67: WyckList = [('a',('0.25','0','0'),4),('b',('0.25','0','0.5'),4),('c',('0','0','0'),4),('d',('0','0','0.5'),4),('e',('0.25','0.25','0'),4),('f',('0.25','0.25','0.5'),4),('g',('0','0.25','z'),4),('h',('x','0','0'),8),('i',('x','0','0.5'),8),('j',('0.25','y','0'),8),('k',('0.25','y','0.5'),8),('l',('0.25','0','z'),8),('m',('0','y','z'),8),('n',('x','0.25','z'),8),('o',('x','y','z'),16)]
	elif SG==68: WyckList = [('a',('0','0.25','0.25'),4),('b',('0','0.25','0.75'),4),('c',('0.25','0.75','0'),8),('d',('0','0','0'),8),('e',('x','0.25','0.25'),8),('f',('0','y','0.25'),8),('g',('0','0.25','z'),8),('h',('0.25','0','z'),8),('i',('x','y','z'),16)]
	elif SG==69: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0','0.25','0.25'),8),('d',('0.25','0','0.25'),8),('e',('0.25','0.25','0'),8),('f',('0.25','0.25','0.25'),8),('g',('x','0','0'),8),('h',('0','y','0'),8),('i',('0','0','z'),8),('j',('0.25','0.25','z'),16),('k',('0.25','y','0.25'),16),('l',('x','0.25','0.25'),16),('m',('0','y','z'),16),('n',('x','0','z'),16),('o',('x','y','0'),16),('p',('x','y','z'),32)]
	elif SG==70: WyckList = [('a',('0.125','0.125','0.125'),8),('b',('0.125','0.125','0.625'),8),('c',('0','0','0'),16),('d',('0.5','0.5','0.5'),16),('e',('x','0.125','0.125'),16),('f',('0.125','y','0.125'),16),('g',('0.125','0.125','z'),16),('h',('x','y','z'),32)]
	elif SG==71: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),2),('c',('0.5','0.5','0'),2),('d',('0.5','0','0.5'),2),('e',('x','0','0'),4),('f',('x','0.5','0'),4),('g',('0','y','0'),4),('h',('0','y','0.5'),4),('i',('0','0','z'),4),('j',('0.5','0','z'),4),('k',('0.25','0.25','0.25'),8),('l',('0','y','z'),8),('m',('x','0','z'),8),('n',('x','y','0'),8),('o',('x','y','z'),16)]
	elif SG==72: WyckList = [('a',('0','0','0.25'),4),('b',('0.5','0','0.25'),4),('c',('0','0','0'),4),('d',('0.5','0','0'),4),('e',('0.25','0.25','0.25'),8),('f',('x','0','0.25'),8),('g',('0','y','0.25'),8),('h',('0','0','z'),8),('i',('0','0.5','z'),8),('j',('x','y','0'),8),('k',('x','y','z'),16)]
	elif SG==73: WyckList = [('a',('0','0','0'),8),('b',('0.25','0.25','0.25'),8),('c',('x','0','0.25'),8),('d',('0.25','y','0'),8),('e',('0','0.25','z'),8),('f',('x','y','z'),16)]
	elif SG==74: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0.25','0.25','0.25'),4),('d',('0.25','0.25','0.75'),4),('e',('0','0.25','z'),4),('f',('x','0','0'),8),('g',('0.25','y','0.25'),8),('h',('0','y','z'),8),('i',('x','0.25','z'),8),('j',('x','y','z'),16)]
	elif SG==75: WyckList = [('a',('0','0','z'),1),('b',('0.5','0.5','z'),1),('c',('0','0.5','z'),2),('d',('x','y','z'),4)]
	elif SG==76: WyckList = [('a',('x','y','z'),4)]
	elif SG==77: WyckList = [('a',('0','0','z'),2),('b',('0.5','0.5','z'),2),('c',('0','0.5','z'),2),('d',('x','y','z'),4)]
	elif SG==78: WyckList = [('a',('x','y','z'),4)]
	elif SG==79: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),4),('c',('x','y','z'),8)]
	elif SG==80: WyckList = [('a',('0','0','z'),4),('b',('x','y','z'),8)]
	elif SG==81: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0.5','0.5','0'),1),('d',('0.5','0.5','0.5'),1),('e',('0','0','z'),2),('f',('0.5','0.5','z'),2),('g',('0','0.5','z'),2),('h',('x','y','z'),4)]
	elif SG==82: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0.25'),2),('d',('0','0.5','0.75'),2),('e',('0','0','z'),4),('f',('0','0.5','z'),4),('g',('x','y','z'),8)]
	elif SG==83: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0.5','0.5','0'),1),('d',('0.5','0.5','0.5'),1),('e',('0','0.5','0'),2),('f',('0','0.5','0.5'),2),('g',('0','0','z'),2),('h',('0.5','0.5','z'),2),('i',('0','0.5','z'),4),('j',('x','y','0'),4),('k',('x','y','0.5'),4),('l',('x','y','z'),8)]
	elif SG==84: WyckList = [('a',('0','0','0'),2),('b',('0.5','0.5','0'),2),('c',('0','0.5','0'),2),('d',('0','0.5','0.5'),2),('e',('0','0','0.25'),2),('f',('0.5','0.5','0.25'),2),('g',('0','0','z'),4),('h',('0.5','0.5','z'),4),('i',('0','0.5','z'),4),('j',('x','y','0'),4),('k',('x','y','z'),8)]
	elif SG==85: WyckList = [('a',('0.25','0.75','0'),2),('b',('0.25','0.75','0.5'),2),('c',('0.25','0.25','z'),2),('d',('0','0','0'),4),('e',('0','0','0.5'),4),('f',('0.25','0.75','z'),4),('g',('x','y','z'),8)]
	elif SG==86: WyckList = [('a',('0.25','0.25','0.25'),2),('b',('0.25','0.25','0.75'),2),('c',('0','0','0'),4),('d',('0','0','0.5'),4),('e',('0.75','0.25','z'),4),('f',('0.25','0.25','z'),4),('g',('x','y','z'),8)]
	elif SG==87: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),4),('f',('0.25','0.25','0.25'),8),('g',('0','0.5','z'),8),('h',('x','y','0'),8),('i',('x','y','z'),16)]
	elif SG==88: WyckList = [('a',('0','0.25','0.125'),4),('b',('0','0.25','0.625'),4),('c',('0','0','0'),8),('d',('0','0','0.5'),8),('e',('0','0.25','z'),8),('f',('x','y','z'),16)]
	elif SG==89: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0.5','0.5','0'),1),('d',('0.5','0.5','0.5'),1),('e',('0.5','0','0'),2),('f',('0.5','0','0.5'),2),('g',('0','0','z'),2),('h',('0.5','0.5','z'),2),('i',('0','0.5','z'),4),('j',('x','x','0'),4),('k',('x','x','0.5'),4),('l',('x','0','0'),4),('m',('x','0.5','0.5'),4),('n',('x','0','0.5'),4),('o',('x','0.5','0'),4),('p',('x','y','z'),8)]
	elif SG==90: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','z'),2),('d',('0','0','z'),4),('e',('x','x','0'),4),('f',('x','x','0.5'),4),('g',('x','y','z'),8)]
	elif SG==91: WyckList = [('a',('0','y','0'),4),('b',('0.5','y','0'),4),('c',('x','x','0.375'),4),('d',('x','y','z'),8)]
	elif SG==92: WyckList = [('a',('x','x','0'),4),('b',('x','y','z'),8)]
	elif SG==93: WyckList = [('a',('0','0','0'),2),('b',('0.5','0.5','0'),2),('c',('0','0.5','0'),2),('d',('0','0.5','0.5'),2),('e',('0','0','0.25'),2),('f',('0.5','0.5','0.25'),2),('g',('0','0','z'),4),('h',('0.5','0.5','z'),4),('i',('0','0.5','z'),4),('j',('x','0','0'),4),('k',('x','0.5','0.5'),4),('l',('x','0','0.5'),4),('m',('x','0.5','0'),4),('n',('x','x','0.25'),4),('o',('x','x','0.75'),4),('p',('x','y','z'),8)]
	elif SG==94: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0','z'),4),('d',('0','0.5','z'),4),('e',('x','x','0'),4),('f',('x','x','0.5'),4),('g',('x','y','z'),8)]
	elif SG==95: WyckList = [('a',('0','y','0'),4),('b',('0.5','y','0'),4),('c',('x','x','0.625'),4),('d',('x','y','z'),8)]
	elif SG==96: WyckList = [('a',('x','x','0'),4),('b',('x','y','z'),8)]
	elif SG==97: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),4),('f',('0','0.5','z'),8),('g',('x','x','0'),8),('h',('x','0','0'),8),('i',('x','0','0.5'),8),('j',('x','x+0.5','0.25'),8),('k',('x','y','z'),16)]
	elif SG==98: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0','0','z'),8),('d',('x','x','0'),8),('e',('-x','x','0'),8),('f',('x','0.25','0.125'),8),('g',('x','y','z'),16)]
	elif SG==99: WyckList = [('a',('0','0','z'),1),('b',('0.5','0.5','z'),1),('c',('0.5','0','z'),2),('d',('x','x','z'),4),('e',('x','0','z'),4),('f',('x','0.5','z'),4),('g',('x','y','z'),8)]
	elif SG==100: WyckList = [('a',('0','0','z'),2),('b',('0.5','0','z'),2),('c',('x','x+0.5','z'),4),('d',('x','y','z'),8)]
	elif SG==101: WyckList = [('a',('0','0','z'),2),('b',('0.5','0.5','z'),2),('c',('0','0.5','z'),4),('d',('x','x','z'),4),('e',('x','y','z'),8)]
	elif SG==102: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),4),('c',('x','x','z'),4),('d',('x','y','z'),8)]
	elif SG==103: WyckList = [('a',('0','0','z'),2),('b',('0.5','0.5','z'),2),('c',('0','0.5','z'),4),('d',('x','y','z'),8)]
	elif SG==104: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),4),('c',('x','y','z'),8)]
	elif SG==105: WyckList = [('a',('0','0','z'),2),('b',('0.5','0.5','z'),2),('c',('0','0.5','z'),2),('d',('x','0','z'),4),('e',('x','0.5','z'),4),('f',('x','y','z'),8)]
	elif SG==106: WyckList = [('a',('0','0','z'),4),('b',('0','0.5','z'),4),('c',('x','y','z'),8)]
	elif SG==107: WyckList = [('a',('0','0','z'),2),('b',('0','0.5','z'),4),('c',('x','x','z'),8),('d',('x','0','z'),8),('e',('x','y','z'),16)]
	elif SG==108: WyckList = [('a',('0','0','z'),4),('b',('0.5','0','z'),4),('c',('x','x+0.5','z'),8),('d',('x','y','z'),16)]
	elif SG==109: WyckList = [('a',('0','0','z'),4),('b',('0','y','z'),8),('c',('x','y','z'),16)]
	elif SG==110: WyckList = [('a',('0','0','z'),8),('b',('x','y','z'),16)]
	elif SG==111: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0.5'),1),('c',('0','0','0.5'),1),('d',('0.5','0.5','0'),1),('e',('0.5','0','0'),2),('f',('0.5','0','0.5'),2),('g',('0','0','z'),2),('h',('0.5','0.5','z'),2),('i',('x','0','0'),4),('j',('x','0.5','0.5'),4),('k',('x','0','0.5'),4),('l',('x','0.5','0'),4),('m',('0','0.5','z'),4),('n',('x','x','z'),4),('o',('x','y','z'),8)]
	elif SG==112: WyckList = [('a',('0','0','0.25'),2),('b',('0.5','0','0.25'),2),('c',('0.5','0.5','0.25'),2),('d',('0','0.5','0.25'),2),('e',('0','0','0'),2),('f',('0.5','0.5','0'),2),('g',('x','0','0.25'),4),('h',('0.5','y','0.25'),4),('i',('x','0.5','0.25'),4),('j',('0','y','0.25'),4),('k',('0','0','z'),4),('l',('0.5','0.5','z'),4),('m',('0','0.5','z'),4),('n',('x','y','z'),8)]
	elif SG==113: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','z'),2),('d',('0','0','z'),4),('e',('x','x+0.5','z'),4),('f',('x','y','z'),8)]
	elif SG==114: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0','z'),4),('d',('0','0.5','z'),4),('e',('x','y','z'),8)]
	elif SG==115: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0'),1),('c',('0.5','0.5','0.5'),1),('d',('0','0','0.5'),1),('e',('0','0','z'),2),('f',('0.5','0.5','z'),2),('g',('0','0.5','z'),2),('h',('x','x','0'),4),('i',('x','x','0.5'),4),('j',('x','0','z'),4),('k',('x','0.5','z'),4),('l',('x','y','z'),8)]
	elif SG==116: WyckList = [('a',('0','0','0.25'),2),('b',('0.5','0.5','0.25'),2),('c',('0','0','0'),2),('d',('0.5','0.5','0'),2),('e',('x','x','0.25'),4),('f',('x','x','0.75'),4),('g',('0','0','z'),4),('h',('0.5','0.5','z'),4),('i',('0','0.5','z'),4),('j',('x','y','z'),8)]
	elif SG==117: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),2),('d',('0','0.5','0.5'),2),('e',('0','0','z'),4),('f',('0','0.5','z'),4),('g',('x','x+0.5','0'),4),('h',('x','x+0.5','0.5'),4),('i',('x','y','z'),8)]
	elif SG==118: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0.25'),2),('d',('0','0.5','0.75'),2),('e',('0','0','z'),4),('f',('x','-x+0.5','0.25'),4),('g',('x','x+0.5','0.25'),4),('h',('0','0.5','z'),4),('i',('x','y','z'),8)]
	elif SG==119: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0.25'),2),('d',('0','0.5','0.75'),2),('e',('0','0','z'),4),('f',('0','0.5','z'),4),('g',('x','x','0'),8),('h',('x','x+0.5','0.25'),8),('i',('x','0','z'),8),('j',('x','y','z'),16)]
	elif SG==120: WyckList = [('a',('0','0','0.25'),4),('b',('0','0','0'),4),('c',('0','0.5','0.25'),4),('d',('0','0.5','0'),4),('e',('x','x','0.25'),8),('f',('0','0','z'),8),('g',('0','0.5','z'),8),('h',('x','x+0.5','0'),8),('i',('x','y','z'),16)]
	elif SG==121: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),4),('f',('x','0','0'),8),('g',('x','0','0.5'),8),('h',('0','0.5','z'),8),('i',('x','x','z'),8),('j',('x','y','z'),16)]
	elif SG==122: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.5'),4),('c',('0','0','z'),8),('d',('x','0.25','0.125'),8),('e',('x','y','z'),16)]
	elif SG==123: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0.5','0.5','0'),1),('d',('0.5','0.5','0.5'),1),('e',('0','0.5','0.5'),2),('f',('0','0.5','0'),2),('g',('0','0','z'),2),('h',('0.5','0.5','z'),2),('i',('0','0.5','z'),4),('j',('x','x','0'),4),('k',('x','x','0.5'),4),('l',('x','0','0'),4),('m',('x','0','0.5'),4),('n',('x','0.5','0'),4),('o',('x','0.5','0.5'),4),('p',('x','y','0'),8),('q',('x','y','0.5'),8),('r',('x','x','z'),8),('s',('x','0','z'),8),('t',('x','0.5','z'),8),('u',('x','y','z'),16)]
	elif SG==124: WyckList = [('a',('0','0','0.25'),2),('b',('0','0','0'),2),('c',('0.5','0.5','0.25'),2),('d',('0.5','0.5','0'),2),('e',('0','0.5','0'),4),('f',('0','0.5','0.25'),4),('g',('0','0','z'),4),('h',('0.5','0.5','z'),4),('i',('0','0.5','z'),8),('j',('x','x','0.25'),8),('k',('x','0','0.25'),8),('l',('x','0.5','0.25'),8),('m',('x','y','0'),8),('n',('x','y','z'),16)]
	elif SG==125: WyckList = [('a',('0.25','0.25','0'),2),('b',('0.25','0.25','0.5'),2),('c',('0.75','0.25','0'),2),('d',('0.75','0.25','0.5'),2),('e',('0','0','0'),4),('f',('0','0','0.5'),4),('g',('0.25','0.25','z'),4),('h',('0.75','0.25','z'),4),('i',('x','x','0'),8),('j',('x','x','0.5'),8),('k',('x','0.25','0'),8),('l',('x','0.25','0.5'),8),('m',('x','-x','z'),8),('n',('x','y','z'),16)]
	elif SG==126: WyckList = [('a',('0.25','0.25','0.25'),2),('b',('0.25','0.25','0.75'),2),('c',('0.25','0.75','0.75'),4),('d',('0.25','0.75','0'),4),('e',('0.25','0.25','z'),4),('f',('0','0','0'),8),('g',('0.25','0.75','z'),8),('h',('x','x','0.25'),8),('i',('x','0.25','0.25'),8),('j',('x','0.75','0.25'),8),('k',('x','y','z'),16)]
	elif SG==127: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0.5'),2),('d',('0','0.5','0'),2),('e',('0','0','z'),4),('f',('0','0.5','z'),4),('g',('x','x+0.5','0'),4),('h',('x','x+0.5','0.5'),4),('i',('x','y','0'),8),('j',('x','y','0.5'),8),('k',('x','x+0.5','z'),8),('l',('x','y','z'),16)]
	elif SG==128: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),4),('f',('0','0.5','z'),8),('g',('x','x+0.5','0.25'),8),('h',('x','y','0'),8),('i',('x','y','z'),16)]
	elif SG==129: WyckList = [('a',('0.75','0.25','0'),2),('b',('0.75','0.25','0.5'),2),('c',('0.25','0.25','z'),2),('d',('0','0','0'),4),('e',('0','0','0.5'),4),('f',('0.75','0.25','z'),4),('g',('x','-x','0'),8),('h',('x','-x','0.5'),8),('i',('0.25','y','z'),8),('j',('x','x','z'),8),('k',('x','y','z'),16)]
	elif SG==130: WyckList = [('a',('0.75','0.25','0.25'),4),('b',('0.75','0.25','0'),4),('c',('0.25','0.25','z'),4),('d',('0','0','0'),8),('e',('0.75','0.25','z'),8),('f',('x','-x','0.25'),8),('g',('x','y','z'),16)]
	elif SG==131: WyckList = [('a',('0','0','0'),2),('b',('0.5','0.5','0'),2),('c',('0','0.5','0'),2),('d',('0','0.5','0.5'),2),('e',('0','0','0.25'),2),('f',('0.5','0.5','0.25'),2),('g',('0','0','z'),4),('h',('0.5','0.5','z'),4),('i',('0','0.5','z'),4),('j',('x','0','0'),4),('k',('x','0.5','0.5'),4),('l',('x','0','0.5'),4),('m',('x','0.5','0'),4),('n',('x','x','0.25'),8),('o',('0','y','z'),8),('p',('0.5','y','z'),8),('q',('x','y','0'),8),('r',('x','y','z'),16)]
	elif SG==132: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.25'),2),('c',('0.5','0.5','0'),2),('d',('0.5','0.5','0.25'),2),('e',('0','0.5','0.25'),4),('f',('0','0.5','0'),4),('g',('0','0','z'),4),('h',('0.5','0.5','z'),4),('i',('x','x','0'),4),('j',('x','x','0.5'),4),('k',('0','0.5','z'),8),('l',('x','0','0.25'),8),('m',('x','0.5','0.25'),8),('n',('x','y','0'),8),('o',('x','x','z'),8),('p',('x','y','z'),16)]
	elif SG==133: WyckList = [('a',('0.25','0.25','0'),4),('b',('0.75','0.25','0'),4),('c',('0.25','0.25','0.25'),4),('d',('0.75','0.25','0.75'),4),('e',('0','0','0'),8),('f',('0.25','0.25','z'),8),('g',('0.75','0.25','z'),8),('h',('x','0.25','0'),8),('i',('x','0.25','0.5'),8),('j',('x','x','0.25'),8),('k',('x','y','z'),16)]
	elif SG==134: WyckList = [('a',('0.25','0.75','0.25'),2),('b',('0.75','0.25','0.25'),2),('c',('0.25','0.25','0.25'),4),('d',('0.25','0.25','0'),4),('e',('0','0','0.5'),4),('f',('0','0','0'),4),('g',('0.75','0.25','z'),4),('h',('0.25','0.25','z'),8),('i',('x','0.25','0.75'),8),('j',('x','0.25','0.25'),8),('k',('x','x','0'),8),('l',('x','x','0.5'),8),('m',('x','-x','z'),8),('n',('x','y','z'),16)]
	elif SG==135: WyckList = [('a',('0','0','0'),4),('b',('0','0','0.25'),4),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),8),('f',('0','0.5','z'),8),('g',('x','x+0.5','0.25'),8),('h',('x','y','0'),8),('i',('x','y','z'),16)]
	elif SG==136: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),4),('f',('x','x','0'),4),('g',('x','-x','0'),4),('h',('0','0.5','z'),8),('i',('x','y','0'),8),('j',('x','x','z'),8),('k',('x','y','z'),16)]
	elif SG==137: WyckList = [('a',('0.75','0.25','0.75'),2),('b',('0.75','0.25','0.25'),2),('c',('0.75','0.25','z'),4),('d',('0.25','0.25','z'),4),('e',('0','0','0'),8),('f',('x','-x','0.25'),8),('g',('0.25','y','z'),8),('h',('x','y','z'),16)]
	elif SG==138: WyckList = [('a',('0.75','0.25','0'),4),('b',('0.75','0.25','0.75'),4),('c',('0','0','0.5'),4),('d',('0','0','0'),4),('e',('0.25','0.25','z'),4),('f',('0.75','0.25','z'),8),('g',('x','-x','0.5'),8),('h',('x','-x','0'),8),('i',('x','x','z'),8),('j',('x','y','z'),16)]
	elif SG==139: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.5'),2),('c',('0','0.5','0'),4),('d',('0','0.5','0.25'),4),('e',('0','0','z'),4),('f',('0.25','0.25','0.25'),8),('g',('0','0.5','z'),8),('h',('x','x','0'),8),('i',('x','0','0'),8),('j',('x','0.5','0'),8),('k',('x','x+0.5','0.25'),16),('l',('x','y','0'),16),('m',('x','x','z'),16),('n',('0','y','z'),16),('o',('x','y','z'),32)]
	elif SG==140: WyckList = [('a',('0','0','0.25'),4),('b',('0','0.5','0.25'),4),('c',('0','0','0'),4),('d',('0','0.5','0'),4),('e',('0.25','0.25','0.25'),8),('f',('0','0','z'),8),('g',('0','0.5','z'),8),('h',('x','x+0.5','0'),8),('i',('x','x','0.25'),16),('j',('x','0','0.25'),16),('k',('x','y','0'),16),('l',('x','x+0.5','z'),16),('m',('x','y','z'),32)]
	elif SG==141: WyckList = [('a',('0','0.75','0.125'),4),('b',('0','0.25','0.375'),4),('c',('0','0','0'),8),('d',('0','0','0.5'),8),('e',('0','0.25','z'),8),('f',('x','0','0'),16),('g',('x','x+0.25','0.875'),16),('h',('0','y','z'),16),('i',('x','y','z'),32)]
	elif SG==142: WyckList = [('a',('0','0.25','0.375'),8),('b',('0','0.25','0.125'),8),('c',('0','0','0'),16),('d',('0','0.25','z'),16),('e',('x','0','0.25'),16),('f',('x','x+0.25','0.125'),16),('g',('x','y','z'),32)]
	elif SG==143: WyckList = [('a',('0','0','z'),1),('b',('1./3.','2./3.','z'),1),('c',('2./3.','1./3.','z'),1),('d',('x','y','z'),3)]
	elif SG==144: WyckList = [('a',('x','y','z'),3)]
	elif SG==145: WyckList = [('a',('x','y','z'),3)]
	elif SG==146: WyckList = [('a',('0','0','z'),3),('b',('x','y','z'),9)]
	elif SG==147: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0','0','z'),2),('d',('1./3.','2./3.','z'),2),('e',('0.5','0','0'),3),('f',('0.5','0','0.5'),3),('g',('x','y','z'),6)]
	elif SG==148: WyckList = [('a',('0','0','0'),3),('b',('0','0','0.5'),3),('c',('0','0','z'),6),('d',('0.5','0','0.5'),9),('e',('0.5','0','0'),9),('f',('x','y','z'),18)]
	elif SG==149: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),1),('d',('1./3.','2./3.','0.5'),1),('e',('2./3.','1./3.','0'),1),('f',('2./3.','1./3.','0.5'),1),('g',('0','0','z'),2),('h',('1./3.','2./3.','z'),2),('i',('2./3.','1./3.','z'),2),('j',('x','-x','0'),3),('k',('x','-x','0.5'),3),('l',('x','y','z'),6)]
	elif SG==150: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0','0','z'),2),('d',('1./3.','2./3.','z'),2),('e',('x','0','0'),3),('f',('x','0','0.5'),3),('g',('x','y','z'),6)]
	elif SG==151: WyckList = [('a',('x','-x','1./3.'),3),('b',('x','-x','5./6.'),3),('c',('x','y','z'),6)]
	elif SG==152: WyckList = [('a',('x','0','1./3.'),3),('b',('x','0','5./6.'),3),('c',('x','y','z'),6)]
	elif SG==153: WyckList = [('a',('x','-x','2./3.'),3),('b',('x','-x','1./6.'),3),('c',('x','y','z'),6)]
	elif SG==154: WyckList = [('a',('x','0','2./3.'),3),('b',('x','0','1./6.'),3),('c',('x','y','z'),6)]
	elif SG==155: WyckList = [('a',('0','0','0'),3),('b',('0','0','0.5'),3),('c',('0','0','z'),6),('d',('x','0','0'),9),('e',('x','0','0.5'),9),('f',('x','y','z'),18)]
	elif SG==156: WyckList = [('a',('0','0','z'),1),('b',('1./3.','2./3.','z'),1),('c',('2./3.','1./3.','z'),1),('d',('x','-x','z'),3),('e',('x','y','z'),6)]
	elif SG==157: WyckList = [('a',('0','0','z'),1),('b',('1./3.','2./3.','z'),2),('c',('x','0','z'),3),('d',('x','y','z'),6)]
	elif SG==158: WyckList = [('a',('0','0','z'),2),('b',('1./3.','2./3.','z'),2),('c',('2./3.','1./3.','z'),2),('d',('x','y','z'),6)]
	elif SG==159: WyckList = [('a',('0','0','z'),2),('b',('1./3.','2./3.','z'),2),('c',('x','y','z'),6)]
	elif SG==160: WyckList = [('a',('0','0','z'),3),('b',('x','-x','z'),9),('c',('x','y','z'),18)]
	elif SG==161: WyckList = [('a',('0','0','z'),6),('b',('x','y','z'),18)]
	elif SG==162: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),2),('d',('1./3.','2./3.','0.5'),2),('e',('0','0','z'),2),('f',('0.5','0','0'),3),('g',('0.5','0','0.5'),3),('h',('1./3.','2./3.','z'),4),('i',('x','-x','0'),6),('j',('x','-x','0.5'),6),('k',('x','0','z'),6),('l',('x','y','z'),12)]
	elif SG==163: WyckList = [('a',('0','0','0.25'),2),('b',('0','0','0'),2),('c',('1./3.','2./3.','0.25'),2),('d',('2./3.','1./3.','0.25'),2),('e',('0','0','z'),4),('f',('1./3.','2./3.','z'),4),('g',('0.5','0','0'),6),('h',('x','-x','0.25'),6),('i',('x','y','z'),12)]
	elif SG==164: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('0','0','z'),2),('d',('1./3.','2./3.','z'),2),('e',('0.5','0','0'),3),('f',('0.5','0','0.5'),3),('g',('x','0','0'),6),('h',('x','0','0.5'),6),('i',('x','-x','z'),6),('j',('x','y','z'),12)]
	elif SG==165: WyckList = [('a',('0','0','0.25'),2),('b',('0','0','0'),2),('c',('0','0','z'),4),('d',('1./3.','2./3.','z'),4),('e',('0.5','0','0'),6),('f',('x','0','0.25'),6),('g',('x','y','z'),12)]
	elif SG==166: WyckList = [('a',('0','0','0'),3),('b',('0','0','0.5'),3),('c',('0','0','z'),6),('d',('0.5','0','0.5'),9),('e',('0.5','0','0'),9),('f',('x','0','0'),18),('g',('x','0','0.5'),18),('h',('x','-x','z'),18),('i',('x','y','z'),36)]
	elif SG==167: WyckList = [('a',('0','0','0.25'),6),('b',('0','0','0'),6),('c',('0','0','z'),12),('d',('0.5','0','0'),18),('e',('x','0','0.25'),18),('f',('x','y','z'),36)]
	elif SG==168: WyckList = [('a',('0','0','z'),1),('b',('1./3.','2./3.','z'),2),('c',('0.5','0','z'),3),('d',('x','y','z'),6)]
	elif SG==169: WyckList = [('a',('x','y','z'),6)]
	elif SG==170: WyckList = [('a',('x','y','z'),6)]
	elif SG==171: WyckList = [('a',('0','0','z'),3),('b',('0.5','0.5','z'),3),('c',('x','y','z'),6)]
	elif SG==172: WyckList = [('a',('0','0','z'),3),('b',('0.5','0.5','z'),3),('c',('x','y','z'),6)]
	elif SG==173: WyckList = [('a',('0','0','z'),2),('b',('1./3.','2./3.','z'),2),('c',('x','y','z'),6)]
	elif SG==174: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),1),('d',('1./3.','2./3.','0.5'),1),('e',('2./3.','1./3.','0'),1),('f',('2./3.','1./3.','0.5'),1),('g',('0','0','z'),2),('h',('1./3.','2./3.','z'),2),('i',('2./3.','1./3.','z'),2),('j',('x','y','0'),3),('k',('x','y','0.5'),3),('l',('x','y','z'),6)]
	elif SG==175: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),2),('d',('1./3.','2./3.','0.5'),2),('e',('0','0','z'),2),('f',('0.5','0','0'),3),('g',('0.5','0','0.5'),3),('h',('1./3.','2./3.','z'),4),('i',('0.5','0','z'),6),('j',('x','y','0'),6),('k',('x','y','0.5'),6),('l',('x','y','z'),12)]
	elif SG==176: WyckList = [('a',('0','0','0.25'),2),('b',('0','0','0'),2),('c',('1./3.','2./3.','0.25'),2),('d',('2./3.','1./3.','0.25'),2),('e',('0','0','z'),4),('f',('1./3.','2./3.','z'),4),('g',('0.5','0','0'),6),('h',('x','y','0.25'),6),('i',('x','y','z'),12)]
	elif SG==177: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),2),('d',('1./3.','2./3.','0.5'),2),('e',('0','0','z'),2),('f',('0.5','0','0'),3),('g',('0.5','0','0.5'),3),('h',('1./3.','2./3.','z'),4),('i',('0.5','0','z'),6),('j',('x','0','0'),6),('k',('x','0','0.5'),6),('l',('x','-x','0'),6),('m',('x','-x','0.5'),6),('n',('x','y','z'),12)]
	elif SG==178: WyckList = [('a',('x','0','0'),6),('b',('x','2*x','0.25'),6),('c',('x','y','z'),12)]
	elif SG==179: WyckList = [('a',('x','0','0'),6),('b',('x','2*x','0.75'),6),('c',('x','y','z'),12)]
	elif SG==180: WyckList = [('a',('0','0','0'),3),('b',('0','0','0.5'),3),('c',('0.5','0','0'),3),('d',('0.5','0','0.5'),3),('e',('0','0','z'),6),('f',('0.5','0','z'),6),('g',('x','0','0'),6),('h',('x','0','0.5'),6),('i',('x','2*x','0'),6),('j',('x','2*x','0.5'),6),('k',('x','y','z'),12)]
	elif SG==181: WyckList = [('a',('0','0','0'),3),('b',('0','0','0.5'),3),('c',('0.5','0','0'),3),('d',('0.5','0','0.5'),3),('e',('0','0','z'),6),('f',('0.5','0','z'),6),('g',('x','0','0'),6),('h',('x','0','0.5'),6),('i',('x','2*x','0'),6),('j',('x','2*x','0.5'),6),('k',('x','y','z'),12)]
	elif SG==182: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.25'),2),('c',('1./3.','2./3.','0.25'),2),('d',('1./3.','2./3.','0.75'),2),('e',('0','0','z'),4),('f',('1./3.','2./3.','z'),4),('g',('x','0','0'),6),('h',('x','2*x','0.25'),6),('i',('x','y','z'),12)]
	elif SG==183: WyckList = [('a',('0','0','z'),1),('b',('1./3.','2./3.','z'),2),('c',('0.5','0','z'),3),('d',('x','0','z'),6),('e',('x','-x','z'),6),('f',('x','y','z'),12)]
	elif SG==184: WyckList = [('a',('0','0','z'),2),('b',('1./3.','2./3.','z'),4),('c',('0.5','0','z'),6),('d',('x','y','z'),12)]
	elif SG==185: WyckList = [('a',('0','0','z'),2),('b',('1./3.','2./3.','z'),4),('c',('x','0','z'),6),('d',('x','y','z'),12)]
	elif SG==186: WyckList = [('a',('0','0','z'),2),('b',('1./3.','2./3.','z'),2),('c',('x','-x','z'),6),('d',('x','y','z'),12)]
	elif SG==187: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),1),('d',('1./3.','2./3.','0.5'),1),('e',('2./3.','1./3.','0'),1),('f',('2./3.','1./3.','0.5'),1),('g',('0','0','z'),2),('h',('1./3.','2./3.','z'),2),('i',('2./3.','1./3.','z'),2),('j',('x','-x','0'),3),('k',('x','-x','0.5'),3),('l',('x','y','0'),6),('m',('x','y','0.5'),6),('n',('x','-x','z'),6),('o',('x','y','z'),12)]
	elif SG==188: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.25'),2),('c',('1./3.','2./3.','0'),2),('d',('1./3.','2./3.','0.25'),2),('e',('2./3.','1./3.','0'),2),('f',('2./3.','1./3.','0.25'),2),('g',('0','0','z'),4),('h',('1./3.','2./3.','z'),4),('i',('2./3.','1./3.','z'),4),('j',('x','-x','0'),6),('k',('x','y','0.25'),6),('l',('x','y','z'),12)]
	elif SG==189: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),2),('d',('1./3.','2./3.','0.5'),2),('e',('0','0','z'),2),('f',('x','0','0'),3),('g',('x','0','0.5'),3),('h',('1./3.','2./3.','z'),4),('i',('x','0','z'),6),('j',('x','y','0'),6),('k',('x','y','0.5'),6),('l',('x','y','z'),12)]
	elif SG==190: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.25'),2),('c',('1./3.','2./3.','0.25'),2),('d',('2./3.','1./3.','0.25'),2),('e',('0','0','z'),4),('f',('1./3.','2./3.','z'),4),('g',('x','0','0'),6),('h',('x','y','0.25'),6),('i',('x','y','z'),12)]
	elif SG==191: WyckList = [('a',('0','0','0'),1),('b',('0','0','0.5'),1),('c',('1./3.','2./3.','0'),2),('d',('1./3.','2./3.','0.5'),2),('e',('0','0','z'),2),('f',('0.5','0','0'),3),('g',('0.5','0','0.5'),3),('h',('1./3.','2./3.','z'),4),('i',('0.5','0','z'),6),('j',('x','0','0'),6),('k',('x','0','0.5'),6),('l',('x','2*x','0'),6),('m',('x','2*x','0.5'),6),('n',('x','0','z'),12),('o',('x','2*x','z'),12),('p',('x','y','0'),12),('q',('x','y','0.5'),12),('r',('x','y','z'),24)]
	elif SG==192: WyckList = [('a',('0','0','0.25'),2),('b',('0','0','0'),2),('c',('1./3.','2./3.','0.25'),4),('d',('1./3.','2./3.','0'),4),('e',('0','0','z'),4),('f',('0.5','0','0.25'),6),('g',('0.5','0','0'),6),('h',('1./3.','2./3.','z'),8),('i',('0.5','0','z'),12),('j',('x','0','0.25'),12),('k',('x','2*x','0.25'),12),('l',('x','y','0'),12),('m',('x','y','z'),24)]
	elif SG==193: WyckList = [('a',('0','0','0.25'),2),('b',('0','0','0'),2),('c',('1./3.','2./3.','0.25'),4),('d',('1./3.','2./3.','0'),4),('e',('0','0','z'),4),('f',('0.5','0','0'),6),('g',('x','0','0.25'),6),('h',('1./3.','2./3.','z'),8),('i',('x','2*x','0'),12),('j',('x','y','0.25'),12),('k',('x','0','z'),12),('l',('x','y','z'),24)]
	elif SG==194: WyckList = [('a',('0','0','0'),2),('b',('0','0','0.25'),2),('c',('1./3.','2./3.','0.25'),2),('d',('1./3.','2./3.','0.75'),2),('e',('0','0','z'),4),('f',('1./3.','2./3.','z'),4),('g',('0.5','0','0'),6),('h',('x','2*x','0.25'),6),('i',('x','0','0'),12),('j',('x','y','0.25'),12),('k',('x','2*x','z'),12),('l',('x','y','z'),24)]
	elif SG==195: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0.5'),1),('c',('0','0.5','0.5'),3),('d',('0.5','0','0'),3),('e',('x','x','x'),4),('f',('x','0','0'),6),('g',('x','0','0.5'),6),('h',('x','0.5','0'),6),('i',('x','0.5','0.5'),6),('j',('x','y','z'),12)]
	elif SG==196: WyckList = [('a',('0','0','0'),4),('b',('0.5','0.5','0.5'),4),('c',('0.25','0.25','0.25'),4),('d',('0.75','0.75','0.75'),4),('e',('x','x','x'),16),('f',('x','0','0'),24),('g',('x','0.25','0.25'),24),('h',('x','y','z'),48)]
	elif SG==197: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('x','x','x'),8),('d',('x','0','0'),12),('e',('x','0.5','0'),12),('f',('x','y','z'),24)]
	elif SG==198: WyckList = [('a',('x','x','x'),4),('b',('x','y','z'),12)]
	elif SG==199: WyckList = [('a',('x','x','x'),8),('b',('x','0','0.25'),12),('c',('x','y','z'),24)]
	elif SG==200: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0.5'),1),('c',('0','0.5','0.5'),3),('d',('0.5','0','0'),3),('e',('x','0','0'),6),('f',('x','0','0.5'),6),('g',('x','0.5','0'),6),('h',('x','0.5','0.5'),6),('i',('x','x','x'),8),('j',('0','y','z'),12),('k',('0.5','y','z'),12),('l',('x','y','z'),24)]
	elif SG==201: WyckList = [('a',('0.25','0.25','0.25'),2),('b',('0','0','0'),4),('c',('0.5','0.5','0.5'),4),('d',('0.25','0.75','0.75'),6),('e',('x','x','x'),8),('f',('x','0.25','0.25'),12),('g',('x','0.75','0.25'),12),('h',('x','y','z'),24)]
	elif SG==202: WyckList = [('a',('0','0','0'),4),('b',('0.5','0.5','0.5'),4),('c',('0.25','0.25','0.25'),8),('d',('0','0.25','0.25'),24),('e',('x','0','0'),24),('f',('x','x','x'),32),('g',('x','0.25','0.25'),48),('h',('0','y','z'),48),('i',('x','y','z'),96)]
	elif SG==203: WyckList = [('a',('0.125','0.125','0.125'),8),('b',('0.625','0.625','0.625'),8),('c',('0','0','0'),16),('d',('0.5','0.5','0.5'),16),('e',('x','x','x'),32),('f',('x','0.125','0.125'),48),('g',('x','y','z'),96)]
	elif SG==204: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('0.25','0.25','0.25'),8),('d',('x','0','0'),12),('e',('x','0','0.5'),12),('f',('x','x','x'),16),('g',('0','y','z'),24),('h',('x','y','z'),48)]
	elif SG==205: WyckList = [('a',('0','0','0'),4),('b',('0.5','0.5','0.5'),4),('c',('x','x','x'),8),('d',('x','y','z'),24)]
	elif SG==206: WyckList = [('a',('0','0','0'),8),('b',('0.25','0.25','0.25'),8),('c',('x','x','x'),16),('d',('x','0','0.25'),24),('e',('x','y','z'),48)]
	elif SG==207: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0.5'),1),('c',('0','0.5','0.5'),3),('d',('0.5','0','0'),3),('e',('x','0','0'),6),('f',('x','0.5','0.5'),6),('g',('x','x','x'),8),('h',('x','0.5','0'),12),('i',('0','y','y'),12),('j',('0.5','y','y'),12),('k',('x','y','z'),24)]
	elif SG==208: WyckList = [('a',('0','0','0'),2),('b',('0.25','0.25','0.25'),4),('c',('0.75','0.75','0.75'),4),('d',('0','0.5','0.5'),6),('e',('0.25','0','0.5'),6),('f',('0.25','0.5','0'),6),('g',('x','x','x'),8),('h',('x','0','0'),12),('i',('x','0','0.5'),12),('j',('x','0.5','0'),12),('k',('0.25','y','-y+0.5'),12),('l',('0.25','y','y+0.5'),12),('m',('x','y','z'),24)]
	elif SG==209: WyckList = [('a',('0','0','0'),4),('b',('0.5','0.5','0.5'),4),('c',('0.25','0.25','0.25'),8),('d',('0','0.25','0.25'),24),('e',('x','0','0'),24),('f',('x','x','x'),32),('g',('0','y','y'),48),('h',('0.5','y','y'),48),('i',('x','0.25','0.25'),48),('j',('x','y','z'),96)]
	elif SG==210: WyckList = [('a',('0','0','0'),8),('b',('0.5','0.5','0.5'),8),('c',('0.125','0.125','0.125'),16),('d',('0.625','0.625','0.625'),16),('e',('x','x','x'),32),('f',('x','0','0'),48),('g',('0.125','y','-y+0.25'),48),('h',('x','y','z'),96)]
	elif SG==211: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('0.25','0.25','0.25'),8),('d',('0.25','0.5','0'),12),('e',('x','0','0'),12),('f',('x','x','x'),16),('g',('x','0.5','0'),24),('h',('0','y','y'),24),('i',('0.25','y','-y+0.5'),24),('j',('x','y','z'),48)]
	elif SG==212: WyckList = [('a',('0.125','0.125','0.125'),4),('b',('0.625','0.625','0.625'),4),('c',('x','x','x'),8),('d',('0.125','y','-y+0.25'),12),('e',('x','y','z'),24)]
	elif SG==213: WyckList = [('a',('0.375','0.375','0.375'),4),('b',('0.875','0.875','0.875'),4),('c',('x','x','x'),8),('d',('0.125','y','y+0.25'),12),('e',('x','y','z'),24)]
	elif SG==214: WyckList = [('a',('0.125','0.125','0.125'),8),('b',('0.875','0.875','0.875'),8),('c',('0.125','0','0.25'),12),('d',('0.625','0','0.25'),12),('e',('x','x','x'),16),('f',('x','0','0.25'),24),('g',('0.125','y','y+0.25'),24),('h',('0.125','y','-y+0.25'),24),('i',('x','y','z'),48)]
	elif SG==215: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0.5'),1),('c',('0','0.5','0.5'),3),('d',('0.5','0','0'),3),('e',('x','x','x'),4),('f',('x','0','0'),6),('g',('x','0.5','0.5'),6),('h',('x','0.5','0'),12),('i',('x','x','z'),12),('j',('x','y','z'),24)]
	elif SG==216: WyckList = [('a',('0','0','0'),4),('b',('0.5','0.5','0.5'),4),('c',('0.25','0.25','0.25'),4),('d',('0.75','0.75','0.75'),4),('e',('x','x','x'),16),('f',('x','0','0'),24),('g',('x','0.25','0.25'),24),('h',('x','x','z'),48),('i',('x','y','z'),96)]
	elif SG==217: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('x','x','x'),8),('d',('0.25','0.5','0'),12),('e',('x','0','0'),12),('f',('x','0.5','0'),24),('g',('x','x','z'),24),('h',('x','y','z'),48)]
	elif SG==218: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('0.25','0.5','0'),6),('d',('0.25','0','0.5'),6),('e',('x','x','x'),8),('f',('x','0','0'),12),('g',('x','0.5','0'),12),('h',('x','0','0.5'),12),('i',('x','y','z'),24)]
	elif SG==219: WyckList = [('a',('0','0','0'),8),('b',('0.25','0.25','0.25'),8),('c',('0','0.25','0.25'),24),('d',('0.25','0','0'),24),('e',('x','x','x'),32),('f',('x','0','0'),48),('g',('x','0.25','0.25'),48),('h',('x','y','z'),96)]
	elif SG==220: WyckList = [('a',('0.375','0','0.25'),12),('b',('0.875','0','0.25'),12),('c',('x','x','x'),16),('d',('x','0','0.25'),24),('e',('x','y','z'),48)]
	elif SG==221: WyckList = [('a',('0','0','0'),1),('b',('0.5','0.5','0.5'),1),('c',('0','0.5','0.5'),3),('d',('0.5','0','0'),3),('e',('x','0','0'),6),('f',('x','0.5','0.5'),6),('g',('x','x','x'),8),('h',('x','0.5','0'),12),('i',('0','y','y'),12),('j',('0.5','y','y'),12),('k',('0','y','z'),24),('l',('0.5','y','z'),24),('m',('x','x','z'),24),('n',('x','y','z'),48)]
	elif SG==222: WyckList = [('a',('0.25','0.25','0.25'),2),('b',('0.75','0.25','0.25'),6),('c',('0','0','0'),8),('d',('0','0.75','0.25'),12),('e',('x','0.25','0.25'),12),('f',('x','x','x'),16),('g',('x','0.75','0.25'),24),('h',('0.25','y','y'),24),('i',('x','y','z'),48)]
	elif SG==223: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('0.25','0','0.5'),6),('d',('0.25','0.5','0'),6),('e',('0.25','0.25','0.25'),8),('f',('x','0','0'),12),('g',('x','0','0.5'),12),('h',('x','0.5','0'),12),('i',('x','x','x'),16),('j',('0.25','y','y+0.5'),24),('k',('0','y','z'),24),('l',('x','y','z'),48)]
	elif SG==224: WyckList = [('a',('0.25','0.25','0.25'),2),('b',('0','0','0'),4),('c',('0.5','0.5','0.5'),4),('d',('0.25','0.75','0.75'),6),('e',('x','x','x'),8),('f',('0.5','0.25','0.75'),12),('g',('x','0.25','0.25'),12),('h',('x','0.25','0.75'),24),('i',('0.5','y','y+0.5'),24),('j',('0.5','y','-y'),24),('k',('x','x','z'),24),('l',('x','y','z'),48)]
	elif SG==225: WyckList = [('a',('0','0','0'),4),('b',('0.5','0.5','0.5'),4),('c',('0.25','0.25','0.25'),8),('d',('0','0.25','0.25'),24),('e',('x','0','0'),24),('f',('x','x','x'),32),('g',('x','0.25','0.25'),48),('h',('0','y','y'),48),('i',('0.5','y','y'),48),('j',('0','y','z'),96),('k',('x','x','z'),96),('l',('x','y','z'),192)]
	elif SG==226: WyckList = [('a',('0.25','0.25','0.25'),8),('b',('0','0','0'),8),('c',('0.25','0','0'),24),('d',('0','0.25','0.25'),24),('e',('x','0','0'),48),('f',('x','0.25','0.25'),48),('g',('x','x','x'),64),('h',('0.25','y','y'),96),('i',('0','y','z'),96),('j',('x','y','z'),192)]
	elif SG==227: WyckList = [('a',('0.125','0.125','0.125'),8),('b',('0.375','0.375','0.375'),8),('c',('0','0','0'),16),('d',('0.5','0.5','0.5'),16),('e',('x','x','x'),32),('f',('x','0.125','0.125'),48),('g',('x','x','z'),96),('h',('0','y','-y'),96),('i',('x','y','z'),192)]
	elif SG==228: WyckList = [('a',('0.125','0.125','0.125'),16),('b',('0.25','0.25','0.25'),32),('c',('0','0','0'),32),('d',('0.875','0.125','0.125'),48),('e',('x','x','x'),64),('f',('x','0.125','0.125'),96),('g',('0.25','y','-y'),96),('h',('x','y','z'),192)]
	elif SG==229: WyckList = [('a',('0','0','0'),2),('b',('0','0.5','0.5'),6),('c',('0.25','0.25','0.25'),8),('d',('0.25','0','0.5'),12),('e',('x','0','0'),12),('f',('x','x','x'),16),('g',('x','0','0.5'),24),('h',('0','y','y'),24),('i',('0.25','y','-y+0.5'),48),('j',('0','y','z'),48),('k',('x','x','z'),48),('l',('x','y','z'),96)]
	elif SG==230: WyckList = [('a',('0','0','0'),16),('b',('0.125','0.125','0.125'),16),('c',('0.125','0','0.25'),24),('d',('0.375','0','0.25'),24),('e',('x','x','x'),32),('f',('x','0','0.25'),48),('g',('0.125','y','-y+0.25'),48),('h',('x','y','z'),96)]

	return WyckList







def zeroOut(val):
	""" sets numbers close enough to zero to really zero """
	if abs(val)<1e-13: return 0.0
	else:	return val


def ZfromLabel(label):
	""" Try to find the atomic number from a label """
	if len(label)<1: return -1
	symb = label[0].upper()
	if len(label) > 1:
		if label[1] in string.ascii_letters:
			symb += label[1].lower()

	try:	Zatom = atomGeneral.baseAtom.symbols.index(symb)
	except:	Zatom = -1
	return Zatom



class simpsonIntegral(object):
	""" calculate a definite integral using adaptive Simpson's rule, only used in DW_factor_M() """

	def __init__(self, f, lo,hi, eps=1e-10, maxIter=2000):
		"""
		Initialize the simpsonIntegral instance.
		f		function to integrate, must be able to just call f(x)
		lo,hi	range to integrate, must have lo <= hi
		eps		tolerance, adjusts convergence
		maxIter	maximum allowed number of iterations

		Example:
			print simpsonIntegral(math.sin,0,math.pi/2).calc()
			# this should return 1
		"""
		self.iter = 0
		try:
			self.lo = float(lo)
			self.hi = float(hi)
			self.eps = float(eps)
			self.maxIter = int(maxIter)
		except:
			raise ValueError('simpsonIntegral(), check values of lo=%r, hi=%r, eps=%r,  maxIter=%r' % (lo,hi,eps,maxIter))

		try:
			vLo = f(self.lo)					# test that f is a function that we can call
			valHi = f(self.hi)
		except:	raise ValueError('simpsonIntegral(), f = "%r" does not appear to be function' % f)
		self.f = f
		self.area = None

		if self.maxIter<=1 or self.eps <=0 or self.hi<self.lo:
			raise ValueError('simpsonIntegral(), check values of lo=%r, hi=%r, eps=%r,  maxIter=%r' % (lo,hi,eps,maxIter))

	def __str__(self):
		""" Return string value for simpsonIntegral. """
		if self.area is None:
			return 'Integral{"%s" [%g, %g]}' % (self.f.__name__,self.lo,self.hi)
		else:
			return 'Integral{"%s" [%g, %g]} = %g' % (self.f.__name__,self.lo,self.hi,self.area)

	def __repr__(self):
		""" Return printable representation for simpsonIntegral. """
		out = 'simpsonIntegral[f=%r, lo=%r, hi=%r, eps=%g, maxIter=%r, area=%r]' % (self.f.__name__,self.lo,self.hi,self.eps,self.maxIter,self.area)
		return out

	def calc(self):
		"Calculate integral of f from a to b with max error of eps."
		if self.lo == self.hi: self.area = 0.0
		self.area = self.recursive_asr(self.lo,self.hi,self.eps,self.simpsons_rule(self.lo,self.hi))
		return self.area

	def simpsons_rule(self,a,b):
		c = (a+b) / 2.0
		h3 = abs(b-a) / 6.0
		return h3*(self.f(a) + 4.0*self.f(c) + self.f(b))

	def recursive_asr(self,a,b,eps,whole):
		"Recursive implementation of adaptive Simpson's rule."
		self.iter += 1
		c = (a+b) / 2.0
		left = self.simpsons_rule(a,c)
		right = self.simpsons_rule(c,b)
		if abs(left + right - whole) <= 15*eps:
			return left + right + (left + right - whole)/15.0
		elif self.iter > self.maxIter:
			return left + right + (left + right - whole)/15.0
		return self.recursive_asr(a,c,eps/2.0,left) + self.recursive_asr(c,b,eps/2.0,right)




if __name__ == '__main__':
	"""
	Main function for LatticeSym.py.

	Test cases for Lattice class to verify correct behavior.
	"""
	try:	testGroup = int(sys.argv[1])
	except:	raise ValueError('./LatticeSym testGroup,  testGroup must be a bit mask, a combo of 1-2048, use -1 for all')
	print 'showing testGroup(s) == %r (= %r)' % (testGroup,bin(testGroup))
	print '\n'


	def make_Silicon():
		atomSi = atomXtal('Si1',0,0,0,WyckoffSymbol='a',DebyeT=645)
		bondSi = bondType('Si1','Si1',  0.23517)
		return Lattice(227, 0.54310206,0,0, 0,0,0, desc='Silicon',Temperature=22.5,alphaT=2.56e-6,atoms=(atomSi,),bonds=(bondSi,))

	def make_GaAs():
		atomGa = atomXtal('Ga001',0,0,0,WyckoffSymbol='a',DebyeT=370)
		atomAs = atomXtal('As001',0.25,0.25,0.25,WyckoffSymbol='c',DebyeT=370.1)
		bondGaAs = bondType('Ga001','As001',0.2448)
		return Lattice(216, 0.56534,0,0, 0,0,0, desc='GaAs',atoms=(atomGa,atomAs),bonds=(bondGaAs,))

	def make_Sapphire():
		atomAl = atomXtal('Al1',0,0,0.3523,valence=+3,DebyeT=1047)
		atomO = atomXtal('O1',0.3064,0,0.25,valence=-2,DebyeT=1047)
		return Lattice(167, 0.4758,0,1.2991, 90,90,120, desc='Al2O3 Sapphire (hexagonal)',atoms=(atomAl,atomO))

	def make_YBCO():
		atomY1 = atomXtal('Y1',0.5, 0.5, 0.5,WyckoffSymbol='h')
		atomBa1 = atomXtal('Ba1',0.5, 0.5, 0.185,WyckoffSymbol='t')
		atomCu1 = atomXtal('Cu1',0.0, 0.0, 0.0,WyckoffSymbol='a')
		atomCu2 = atomXtal('Cu2',0.0, 0.0, 0.3565,WyckoffSymbol='q')
		atomO1 = atomXtal('O1',0.0, 0.5, 0.0, WyckoffSymbol='e')
		atomO2 = atomXtal('O2',0.0, 0.0, 0.1566, WyckoffSymbol='q')
		atomO3 = atomXtal('O3',0.5, 0.0, 0.3776, WyckoffSymbol='s')
		atomO4 = atomXtal('O4',0.0, 0.5, 0.3765, WyckoffSymbol='r')
		bond1 = bondType('Y1','O3', 0.2418)
		bond2 = bondType('Y1','O4', 0.2399)
		bond3 = bondType('Ba1','O1', 0.2891)
		bond4 = bondType('Ba1','O2', 0.275)
		bond5 = bondType('Ba1','O3', 0.298)
		bond6 = bondType('Ba1','O4', 0.2948)
		bond7 = bondType('Cu1','O1', 0.1947)
		bond8 = bondType('Cu1','O2', 0.1834)
		bond9 = bondType('Cu2','O2', 0.2341)
		bond10 = bondType('Cu2','O3', 0.1929)
		bond11 = bondType('Cu2','O4', 0.1961)
		atoms = (atomY1,atomBa1,atomCu1,atomCu2,atomO1,atomO2,atomO3,atomO4)
		bonds = (bond1,bond2,bond3,bond4,bond5,bond6,bond7,bond8,bond9,bond10,bond11)
		return Lattice(47, 0.3827,0.3893,1.1699, 90,90,90, desc='YBa2Cu3O7',atoms=atoms,bonds=bonds)

	silicon = make_Silicon()
	GaAs = make_GaAs()
	sapphire = make_Sapphire()
	YBCO = make_YBCO()

	if (4+8+16+32+64) & testGroup:
		print silicon	# Silicon
		print '\n\n=============================================================================\n\n'
	if (4+8+16+512) & testGroup:
		print GaAs		# GaAs
		print '\n\n=============================================================================\n\n'
	if (4+8+32) & testGroup:
		print sapphire	# Sapphire
		print '\n\n=============================================================================\n\n'
	if (8+32+256+1024) & testGroup:
		print YBCO		# YBCO
	print ' '


	def test_FstructAllow(xtal,h,k,l):
		if xtal.allowedHKL(h,k,l):	str = "    Allowed"
		else:						str = "Not Allowed"
		print '%s  F("%s", %g,%g,%g) = %r,   d=%g nm' % (str,xtal.desc,h,k,l,xtal.Fstruct(h,k,l),xtal.dSpacing(h,k,l))

	def test_SymString2SGs(sin,type):
		print 'SymString2SGs(%s, %g) = %r' % (sin,type, SymString2SGs(sin,type))

	def test_angleBetweenHKLs(xtal, h1,k1,l1, h2,k2,l2):
		print 'for "%s", (%g,%g,%g) ^ (%g,%g,%g) = %g°' % (xtal.desc,h1,k1,l1, h2,k2,l2,xtal.angleBetweenHKLs(h1,k1,l1, h2,k2,l2))

	def test_findClosestHKL(xtal,d):
		l = xtal.findClosestHKL(d)
		print 'in "%s", closest d-spacing to %g, there are %d close:' % (xtal.desc,d,len(l))
		for d in l: print d

	def test_isRhombohedral(SG):
		print 'isRhombohedral(%r) --> %r' % (SG, isRhombohedral(SG))

	def test_DW_factor_M(xtal,T,fromIgor):
		DW = xtal.DW_factor_M(T,645.0,20)
		print 'DW_factor_M(%g) = %r,  delta=%g' % (T, DW,DW-fromIgor)

	def test_FindWyckoffSymbol(xtal,x0,y0,z0):
		(symbol,mult) = FindWyckoffSymbol(xtal.SpaceGroup,x0,y0,z0)
		return 'Find(%s, %g,%g,%g) --> Wyckoff = "%s",  mult = %g\t' % (xtal.desc,x0,y0,z0,symbol,mult)

	def test_ForceXYZtoWyckoff(xtal,symbol,x0,y0,z0):
		out = '%s, Wyckoff="%s",   {%g, %g, %g} --> ' % (xtal.desc,symbol,x0,y0,z0)
		(x,y,z) = ForceXYZtoWyckoff(xtal.SpaceGroup,symbol,x0,y0,z0)
		out += "{%g, %g, %g}" % (x,y,z)
		return out


	if testGroup & 1:
		print '\n---testGroup 1'
		test_SymString2SGs('P63*',1)
		test_SymString2SGs('P63*',0)
		test_SymString2SGs('P63*2',0)
		test_SymString2SGs('P1*',0)
		test_SymString2SGs('*2b*',3)
		test_SymString2SGs('*2b*',1)
		test_SymString2SGs('*2b*',0)
		test_SymString2SGs('Tri*c',4)

	if testGroup & 2:
		print '\n---testGroup 2'
		test_FstructAllow(silicon,0,0,0)
		test_FstructAllow(silicon,1,1,1)
		test_FstructAllow(silicon,2,2,2)
		test_FstructAllow(silicon,4,4,2)
		test_FstructAllow(silicon,1,2,3)
		test_FstructAllow(silicon,2,2,0)
		test_FstructAllow(silicon,0,0,4)
		print ' '
		test_FstructAllow(GaAs,1,1,1)
		test_FstructAllow(GaAs,2,2,2)
		test_FstructAllow(GaAs,4,4,2)
		test_FstructAllow(GaAs,1,2,3)
		test_FstructAllow(GaAs,2,2,0)
		test_FstructAllow(GaAs,0,0,4)
		print ' '
		test_FstructAllow(sapphire,0,0,3)
		test_FstructAllow(sapphire,0,0,5)
		test_FstructAllow(sapphire,0,0,6)

	if testGroup & 4:
		print '\n---testGroup 4'
		test_angleBetweenHKLs(silicon,1,0,0, 2,0,2)
		test_angleBetweenHKLs(silicon,1,0,0, 0,2,2)
		test_angleBetweenHKLs(silicon,1,1,1, 2,0,2)
		print ' '
		test_angleBetweenHKLs(GaAs,1,0,0, 2,0,2)
		test_angleBetweenHKLs(GaAs,1,0,0, 0,2,2)
		test_angleBetweenHKLs(GaAs,1,1,1, 2,0,2)
		print ' '
		test_angleBetweenHKLs(YBCO,1,0,0, 2,0,2)
		test_angleBetweenHKLs(YBCO,1,0,0, 0,2,2)
		test_angleBetweenHKLs(YBCO,1,1,1, 2,0,2)
		print ' '
		test_angleBetweenHKLs(sapphire,1,0,0, 2,0,2)
		test_angleBetweenHKLs(sapphire,1,0,0, 0,2,2)
		test_angleBetweenHKLs(sapphire,1,1,1, 2,0,2)

	if testGroup & 8:
		print '\n---testGroup 8'
		test_findClosestHKL(silicon,0.23)
		test_findClosestHKL(GaAs,0.23)

	if testGroup & 16:
		print '\n---testGroup 16'
		test_isRhombohedral(sapphire.SpaceGroup)
		test_isRhombohedral(silicon.SpaceGroup)
		test_isRhombohedral(YBCO.SpaceGroup)
		test_isRhombohedral(None)

	if testGroup & 32:
		print '\n---testGroup 32'
		test_DW_factor_M(silicon.atoms[0],645,0.016508709)
		test_DW_factor_M(silicon.atoms[0],700,0.0178446)
		test_DW_factor_M(silicon.atoms[0],1,0.00401676310870523)
		test_DW_factor_M(silicon.atoms[0],2000,0.049963311)

	if testGroup & 64:
		print '\n---testGroup 64'
		iii = simpsonIntegral(math.sin,0,math.pi/2)	# init values for integration
		print iii				# print, but note there is no value
		iii.calc()				# calculate the integral
		print iii				# print again, note that there is a value this time
		print '%r' % iii		# print the representation
		print 'Integral(sin) over [0,PI/2] = ',simpsonIntegral(math.sin,0,math.pi/2).calc()

	if testGroup & 128:
		print '\n---testGroup 128'
		print ' '
		print ' '
		print ' *************** '
		print '  '
#		YBCOfile = Lattice(0, 0,0,0,0,0,0,file='/Users/tischler/Dropbox/Lattice python/YBCO.xml')
		YBCOfile = Lattice(0, 0,0,0,0,0,0,file='test/YBCO.xml')
		print YBCOfile

	if testGroup & 256:
		print '\n---testGroup 256'
		print ' '
		print ' '
		print ' *************** '
		print '  '
		GaAsfile = Lattice(0, 0,0,0,0,0,0,file='test/GaAs.xml')
		print GaAsfile

	if testGroup & 512:
		print '\n---testGroup 512'
		print ' '
		print ' *************'
		print ' '
		print test_FindWyckoffSymbol(YBCO, 0, 0.5, 0.3765)+'\tWyckoff = "r",  mult = 2'
		print test_FindWyckoffSymbol(YBCO, 0.11, 0.5, 0.3765)+'Wyckoff = "x",  mult = 4'
		print test_FindWyckoffSymbol(YBCO, 0.11, 0.2, 0.3765)+'Wyckoff = "A",  mult = 8'
		print test_FindWyckoffSymbol(YBCO, 0,0,0)+'\t\t\tWyckoff = "a",  mult = 1'
		print test_FindWyckoffSymbol(YBCO, float('nan'),float('nan'),float('nan'))+'\tWyckoff = " ",  mult = 0'

	if testGroup & 1024:
		print '\n---testGroup 1024'
		print ' '
		print ' *************'
		print ' '
		print test_ForceXYZtoWyckoff(YBCO,'A',0.1,0.2,0.3)+"\t#47, Wyckoff='A',   {0.1, 0.2, 0.3} --> {0.1, 0.2, 0.3}"
		print test_ForceXYZtoWyckoff(YBCO,'a',0.1,0.2,0.3)+"\t\t\t#47, Wyckoff='a',   {0.1, 0.2, 0.3} --> {0, 0, 0}"
		print test_ForceXYZtoWyckoff(YBCO,'j',0.1,0.2,0.3)+"\t\t#47, Wyckoff='j',   {0.1, 0.2, 0.3} --> {0.1, 0, 0.5}"

	if testGroup & 2048:
		print '\n---testGroup 2048'
		print ' '
		print ' '
		print ' *************** '
		print '  '
		PuAlpha = Lattice(0, 0,0,0,0,0,0,file='materials/Pu-alpha.xml')
		print PuAlpha

	#		add tests for
	#		def Hex2Rhom(self,aH,cH)			# convert Hexagonal lattice constants to Rhombohedral
	#		def Rhom2Hex(self,aR,alpha)			# convert Rhombohedral lattice constants to Hexagonal





